import math
import random
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
from reil import stateful
from scipy import stats


def rlnormRestricted(meanVal: float, stdev: float) -> float:
    # capture 50% of the data.  This restricts the log values to a "reasonable" range
    quartileRange = (0.25, 0.75)
    lnorm = stats.lognorm(stdev, scale=math.exp(meanVal))
    qValues = lnorm.ppf(quartileRange)
    values = list(v for v in lnorm.rvs(size=1000)
                  if (v > qValues[0]) & (v < qValues[1]))
    return random.sample(values, 1)[0]


class Patient(stateful.Stateful):
    '''
    Two compartment PK/PD model for wafarin.

    Attributes
    ----------
        dose: a dataframe containing with day as key and dose as value.

    Methods
    -------
        INR: returns a list of INR values corresponding to the given list of days.
    '''

    def __init__(self, age: float = 50.0, CYP2C9: str = '*1/*1', VKORC1: str = 'G/A',
                 randomized: bool = True, max_time: int = 24,
                 dose_interval: int = 24, dose: Optional[dict] = None, lazy: bool = False,
                 **kwargs: Any):

        super().__init__(name=kwargs.get('name', __name__),
                         logger_name=kwargs.get('logger_name', __name__),
                         **kwargs)

        self._age = age
        self._CYP2C9 = CYP2C9
        self._VKORC1 = VKORC1
        self._randomized = randomized
        self._MTT_1 = float(kwargs.get('MTT_1', rlnormRestricted(
            math.log(11.6), math.sqrt(0.141)) if randomized else math.log(11.6)))
        self._MTT_2 = float(kwargs.get('MTT_2', rlnormRestricted(
            math.log(120.0), math.sqrt(1.02)) if randomized else math.log(120.0)))

        self._gamma = 0.424  # no units

        # EC_50 in mg/L
        EC_50 = kwargs.get('EC_50', None)
        if EC_50 is None:
            if VKORC1 == "G/G":  # Order of genotypes changed
                EC_50 = rlnormRestricted(
                    math.log(4.61), math.sqrt(0.409)) if randomized else math.log(4.61)
            elif VKORC1 in ["G/A", "A/G"]:
                EC_50 = rlnormRestricted(
                    math.log(3.02), math.sqrt(0.409)) if randomized else math.log(3.02)
            elif VKORC1 == "A/A":
                EC_50 = rlnormRestricted(
                    math.log(2.20), math.sqrt(0.409)) if randomized else math.log(2.20)
            else:
                raise ValueError('The VKORC1 genotype is not supported!')

        self._EC_50_gamma = EC_50 ** self._gamma

        self._cyp_1_1: float = kwargs.get('cyp_1_1', rlnormRestricted(
            math.log(0.314), math.sqrt(0.31)) if randomized else math.log(0.314))
        self._V1: float = kwargs.get('V1', rlnormRestricted(
            math.log(13.8), math.sqrt(0.262)) if randomized else math.log(13.8))
        self._V2: float = kwargs.get('V2', rlnormRestricted(
            math.log(6.59), math.sqrt(0.991)) if randomized else math.log(6.59))
        self._Q = 0.131    # (L/h)
        self._lambda = 3.61

        # bioavilability fraction 0-1 (from: "Applied Pharmacokinetics & Pharmacodynamics 4th edition, p.717", some other references)
        self._F = 0.9

        self._ka = 2  # absorption rate (1/hr)

        ktr1 = 6/self._MTT_1					# 1/hours; changed from 1/MTT_1
        ktr2 = 1/self._MTT_2					# 1/hours
        self._ktr = np.array([0.0] + [ktr1] * 6 + [0.0, ktr2])
        self._E_MAX = 1					        	# no units

        self._CL_s = 1
        if self._age > 71:
            self._CL_s = 1 - 0.0091 * (self._age - 71)

        if self._CYP2C9 == "*1/*1":
            self._CL_s = self._CL_s * self._cyp_1_1
        elif self._CYP2C9 == "*1/*2":
            self._CL_s = self._CL_s * self._cyp_1_1 * (1 - 0.315)
        elif self._CYP2C9 == "*1/*3":
            self._CL_s = self._CL_s * self._cyp_1_1 * (1 - 0.453)
        elif self._CYP2C9 == "*2/*2":
            self._CL_s = self._CL_s * self._cyp_1_1 * (1 - 0.722)
        elif self._CYP2C9 == "*2/*3":
            self._CL_s = self._CL_s * self._cyp_1_1 * (1 - 0.69)
        elif self._CYP2C9 == "*3/*3":
            self._CL_s = self._CL_s * self._cyp_1_1 * (1 - 0.852)
        else:
            raise ValueError('The CYP2C9 genotype not recognized fool!')

        self._max_time = max_time  # The last hour of experiment
        self._dose_interval = dose_interval
        self._dose = dose if dose is not None else {}
        self._lazy = lazy
        self._last_computed_day = 1

        # prepend time 0 to the list of times for deSolve initial conditions (remove when returning list of times)
        len_times = self._max_time + 1
        times = list(range(len_times))
        # times also equals the time-step for deSolve

        k12 = self._Q / self._V1
        k21 = self._Q / self._V2
        k10 = self._CL_s / self._V1
        b = k10 + k21 + k12
        c = k10 * k21
        alpha = (b + math.sqrt(b ** 2 - 4*c)) / 2
        beta = (b - math.sqrt(b ** 2 - 4*c)) / 2

        # 2-compartment model
        part_1 = np.array(list(((k21 - alpha) / ((self._ka - alpha)*(beta - alpha)))
                               * temp for temp in (math.exp(-alpha * t) for t in times)))
        part_2 = np.array(list(((k21 - beta) / ((self._ka - beta)*(alpha - beta)))
                               * temp for temp in (math.exp(-beta * t) for t in times)))
        part_3 = np.array(list(((k21 - self._ka) / ((self._ka - alpha)*(self._ka - beta)))
                               * temp for temp in (math.exp(-self._ka * t) for t in times)))

        multiplication_term = ((self._ka * self._F / 2) /
                               self._V1) * (part_1 + part_2 + part_3).clip(min=1e-5)

        if self._randomized:
            self._Cs_error = np.exp(np.random.normal(0, 0.09, len_times))
            self._Cs_error_ss = np.exp(np.random.normal(0, 0.30, len_times))
            self._exp_e_INR = np.exp(np.random.normal(0, 0.0325, len_times))
        else:
            self._Cs_error = self._Cs_error_ss = np.ones(len_times)
            self._exp_e_INR = np.ones(len_times)

        if self._lazy:
            self.dose = {0: 0.0}
            self._Cs_values = pd.DataFrame(columns=times)
        else:
            self.dose = {0: 0.0}
            self._total_Cs = np.zeros(len_times)
            multiplication_term = np.trim_zeros(multiplication_term, 'b')

        self._multiplication_term_err = np.multiply(
            multiplication_term, self._Cs_error).clip(min=0)
        self._multiplication_term_err_ss = np.multiply(
            multiplication_term, self._Cs_error_ss).clip(min=0)

    @property
    def dose(self) -> Dict[int, float]:
        return self._dose

    @dose.setter
    def dose(self, dose: Dict[int, float]) -> None:
        if self._lazy:
            for day, v in dose.items():
                if v != 0.0:
                    self._Cs_values.loc[day] = self._Cs(
                        dose=v, t0=day * self._dose_interval)
                    self._dose[day] = v
                    self._last_computed_day = min(day, self._last_computed_day)
        else:
            for day, v in dose.items():
                if v != 0.0:
                    self._dose[day] = v
                    range_start = day * self._dose_interval
                    Cs = self._Cs(dose=v, t0=range_start, zero_padding=False)
                    range_end = min(
                        range_start + Cs.shape[0], self._max_time + 1)
                    self._total_Cs[range_start:range_end] += Cs[:range_end-range_start]
                    self._last_computed_day = min(day, self._last_computed_day)

    def _Cs(self, dose: float, t0: int, zero_padding: bool = True) -> np.array:
        # C_s_pred = dose * self._multiplication_term

        # if t0 == 0:  # non steady-state
        #     C_s_error = np.array([exp(gauss(0, 0.09)) for _ in range(len(C_s_pred))]) if self._randomized \
        #         else np.ones(len(C_s_pred))  # Sadjad
        # else:  # steady-state
        #     C_s_error = np.array([exp(gauss(0, 0.30)) for _ in range(len(C_s_pred))]) if self._randomized \
        #         else np.ones(len(C_s_pred))

        # C_s = np.multiply(C_s_pred, C_s_error).clip(min=0)

        # return np.pad(C_s, (t0, 0), 'constant', constant_values=(0,))[:self._max_time+1]

        if t0 == 0:
            C_s = dose * self._multiplication_term_err
        else:
            C_s = dose * self._multiplication_term_err_ss

        return np.concatenate((np.array([0]*t0), C_s[:-t0])) if zero_padding else C_s

    def INR(self, days: Union[int, List[int]]) -> List[float]:
        if isinstance(days, int):
            days = [days]

        if self._lazy:
            base_term = np.sum(self._Cs_values[list(
                range(int(max(days)*self._dose_interval)+1))], axis=0)
        else:
            base_term = self._total_Cs[0:int(max(days)*self._dose_interval)+1]

        Cs_gamma = np.power(
            base_term, self._gamma)

        start_days = sorted(
            [0 if days[0] < self._last_computed_day else self._last_computed_day] + days[:-1])
        end_days = sorted(days)

        if start_days[0] == 0:
            self._A = np.array([0.0] + [1.0] * 8)
            # self._A = [1]*7
            # self._dA = [0]*7

        INR_max = 20
        baseINR = 1
        INR = []
        for d1, d2 in zip(start_days, end_days):
            for i in range(int(d1*self._dose_interval), int(d2*self._dose_interval)):
                inflow = 1 - self._E_MAX * Cs_gamma[i] / (self._EC_50_gamma + Cs_gamma[i])
                self._A[0] = inflow
                self._A[7] = inflow
                self._A[1:] += self._ktr[1:] * (self._A[0:-1] - self._A[1:])

                # self._dA[0] = self._ktr1 * (1 - self._E_MAX * Cs_gamma[i] /
                #                             (self._EC_50 ** self._gamma + Cs_gamma[i])) - self._ktr1*self._A[0]
                # for j in range(1, 6):
                #     self._dA[j] = self._ktr1 * (self._A[j-1] - self._A[j])

                # self._dA[6] = self._ktr2 * (1 - self._E_MAX * Cs_gamma[i] /
                #                             (self._EC_50 ** self._gamma + Cs_gamma[i])) - self._ktr2*self._A[6]
                # for j in range(7):
                #     self._A[j] += self._dA[j]

            INR.append(
                (baseINR + (INR_max*(1-self._A[6]*self._A[8]) ** self._lambda)) * self._exp_e_INR[d2])
            # INR.append(
            #     (baseINR + (INR_max*(1-self._A[5]*self._A[6]) ** self._lambda)) * self._exp_e_INR[d2])

        self._last_computed_day = end_days[-1]

        return INR


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from time import time

    max_day = 100
    p_count = range(20)
    p = [Patient(randomized=False, max_time=24*max_day + 1) for _ in p_count]
    t = time()
    for j in p_count:
        p[j].dose = {i: 15 for i in range(max_day)}
        # plt.plot(p.INR(list(i/24 for i in range(1, max_day*24 + 1))))
        plt.plot(list(range(24, (max_day+1)*24, 240)),
                 p[j].INR(list(i for i in range(1, max_day+1, 10))), 'x')
    plt.show()

    # p = [Patient(randomized=True, max_time=24*max_day + 1) for _ in p_count]
    # t = time()
    # for j in p_count:
    #     p[j].dose = {i: 15 for i in range(max_day)}
    #     p[j].INR(list(i for i in range(1, max_day+1, 10)))
    #     # plt.plot(p.INR(list(i/24 for i in range(1, max_day*24 + 1))))
    # #     plt.plot(list(range(24, (max_day+1)*24, 240)),
    # #              p[j].INR(list(i for i in range(1, max_day+1, 10))), 'x')
    # # plt.show()
    # print(time() - t)

    # p = [Patient(randomized=True, max_time=24*max_day + 1) for _ in p_count]
    # t = time()
    # for j in p_count:
    #     p[j].dose = {i: 15 for i in range(max_day)}
    #     for i in range(max_day, 1, -10):
    #         p[j].INR(i)

    # print(time() - t)



import math
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
from reil.subjects.healthcare.mathematical_models import HealthMathModel


class HambergPKPDFixedMaxTime(HealthMathModel):
    def __init__(self,
                 randomized: bool = True,
                 max_time: int = 24,
                 dose_interval: int = 24,
                 lazy: bool = False) -> None:
        self._randomized = randomized
        self._max_time = max_time  # The last hour of experiment
        self._dose_interval = dose_interval
        self._lazy = lazy

        self._gamma = 0.424  # no units

        self._Q = 0.131    # (L/h)
        self._lambda = 3.61

        # bioavilability fraction 0-1 (from: "Applied Pharmacokinetics &
        # Pharmacodynamics 4th edition, p.717", some other references)
        self._F = 0.9

        self._ka = 2  # absorption rate (1/hr)
        self._E_MAX = 1	 # no units

        self._dose = {}

    def setup(self, **arguments):
        self._age: float = arguments['age'].value  # type:ignore
        self._CYP2C9 = str(arguments['CYP2C9'].value)
        self._VKORC1 = str(arguments['VKORC1'].value)

        self._MTT_1: float = arguments['MTT_1'].value  # type:ignore
        self._MTT_2: float = arguments['MTT_2'].value  # type:ignore

        self._EC_50_gamma: float = arguments['EC_50'].value ** self._gamma

        self._cyp_1_1: float = arguments['cyp_1_1'].value  # type:ignore
        self._V1: float = arguments['V1'].value  # type:ignore
        self._V2: float = arguments['V2'].value  # type:ignore

        ktr1 = 6/self._MTT_1  # type:ignore  # 1/hours; changed from 1/MTT_1
        ktr2 = 1/self._MTT_2  # type:ignore  # 1/hours
        self._ktr = np.array([0.0] + [ktr1] * 6 + [0.0, ktr2])

        temp = {'*1/*1': 0.0,
                '*1/*2': 0.315,
                '*1/*3': 0.453,
                '*2/*2': 0.722,
                '*2/*3': 0.69,
                '*3/*3': 0.852}

        if self._CYP2C9 not in temp:
            raise ValueError('The CYP2C9 genotype not recognized fool!')

        self._CL_s = 1.0 - (0.0091 * (self._age - 71) if self._age > 71 else 0)
        self._CL_s *= self._cyp_1_1 * (1 - temp[self._CYP2C9])

        self._last_computed_day = 1
        self._A = np.array([0.0] + [1.0] * 8)

        # prepend time 0 to the list of times for deSolve
        # initial conditions (remove when returning list of times)
        len_times = self._max_time + 1
        times = list(range(len_times))
        # times also equals the time-step for deSolve

        k12 = self._Q / self._V1  # type:ignore
        k21 = self._Q / self._V2  # type:ignore
        k10 = self._CL_s / self._V1
        b = k10 + k21 + k12
        c = k10 * k21
        alpha = (b + math.sqrt(b ** 2 - 4*c)) / 2
        beta = (b - math.sqrt(b ** 2 - 4*c)) / 2

        # 2-compartment model
        part_1 = np.array(
            list(((k21 - alpha) / ((self._ka - alpha)*(beta - alpha)))
                 * temp for temp in (math.exp(-alpha * t) for t in times)))
        part_2 = np.array(
            list(((k21 - beta) / ((self._ka - beta)*(alpha - beta)))
                 * temp for temp in (math.exp(-beta * t) for t in times)))
        part_3 = np.array(
            list(((k21 - self._ka) / ((self._ka - alpha)*(self._ka - beta)))
                 * temp for temp in (math.exp(-self._ka * t) for t in times)))

        multiplication_term = (
            (self._ka * self._F / 2) /
            self._V1) * (part_1 + part_2 + part_3).clip(min=1e-5)

        if self._randomized:
            self._Cs_error = np.exp(  # type:ignore
                np.random.normal(0, 0.09, len_times))
            self._Cs_error_ss = np.exp(  # type:ignore
                np.random.normal(0, 0.30, len_times))
            self._exp_e_INR = np.exp(  # type:ignore
                np.random.normal(0, 0.0325, len_times))
        else:
            self._Cs_error = self._Cs_error_ss = np.ones(len_times)
            self._exp_e_INR = np.ones(len_times)

        self.dose = {0: 0.0}
        if self._lazy:
            self._Cs_values = pd.DataFrame(columns=times)
        else:
            self._total_Cs = np.zeros(len_times)
            multiplication_term = np.trim_zeros(multiplication_term, 'b')

        self._multiplication_term_err = np.multiply(  # type:ignore
            multiplication_term, self._Cs_error).clip(min=0)
        self._multiplication_term_err_ss = np.multiply(  # type:ignore
            multiplication_term, self._Cs_error_ss).clip(min=0)

    def run(self, **inputs) -> Dict[str, Any]:
        '''
        inputs should include:
        - a dictionary called "dose" with days for each dose as keys and
          the amount of dose as values.
        - a list called "measurement_days" that shows INRs of which days
          should be returned.
        '''
        self.dose = inputs.get('dose', {})

        return {'INR': self.INR(inputs.get('measurement_days', []))}

    @property
    def dose(self) -> Dict[int, float]:
        return self._dose

    @dose.setter
    def dose(self, dose: Dict[int, float]) -> None:
        if self._lazy:
            for day, v in dose.items():
                if v != 0.0:
                    self._Cs_values.loc[day] = self._Cs(
                        dose=v, t0=day * self._dose_interval)
                    self._dose[day] = v
                    self._last_computed_day = min(day, self._last_computed_day)
        else:
            for day, v in dose.items():
                if v != 0.0:
                    self._dose[day] = v
                    range_start = day * self._dose_interval
                    Cs = self._Cs(dose=v, t0=range_start, zero_padding=False)
                    range_end = min(
                        range_start + Cs.shape[0], self._max_time + 1)
                    self._total_Cs[range_start:range_end] += \
                        Cs[:range_end-range_start]
                    self._last_computed_day = min(day, self._last_computed_day)

    def _Cs(self, dose: float, t0: int, zero_padding: bool = True) -> np.array:
        # C_s_pred = dose * self._multiplication_term

        # if t0 == 0:  # non steady-state
        #     C_s_error = np.array([exp(gauss(0, 0.09))
        #                           for _ in range(len(C_s_pred))]) \
        #                   if self._randomized \
        #         else np.ones(len(C_s_pred))  # Sadjad
        # else:  # steady-state
        #     C_s_error = np.array([exp(gauss(0, 0.30)) 
        #                           for _ in range(len(C_s_pred))]) \
        #                   if self._randomized \
        #         else np.ones(len(C_s_pred))

        # C_s = np.multiply(C_s_pred, C_s_error).clip(min=0)

        # return np.pad(C_s, (t0, 0), 'constant', 
        #     constant_values=(0,))[:self._max_time+1]

        if t0 == 0:
            C_s = dose * self._multiplication_term_err
        else:
            C_s = dose * self._multiplication_term_err_ss

        return np.concatenate(
            (np.array([0]*t0), C_s[:-t0])) if zero_padding else C_s

    def INR(self, days: Union[int, List[int]]) -> List[float]:
        if isinstance(days, int):
            days = [days]

        if self._lazy:
            base_term = np.sum(self._Cs_values[list(
                range(int(max(days)*self._dose_interval)+1))], axis=0)
        else:
            base_term = self._total_Cs[0:int(max(days)*self._dose_interval)+1]

        Cs_gamma = np.power(base_term, self._gamma)  # type:ignore

        start_days = sorted([0
                             if days[0] < self._last_computed_day
                             else
                             self._last_computed_day] + days[:-1])

        end_days = sorted(days)

        if start_days[0] == 0:
            self._A = np.array([0.0] + [1.0] * 8)
            # self._A = [1]*7
            # self._dA = [0]*7

        INR_max = 20
        baseINR = 1
        INR = []
        for d1, d2 in zip(start_days, end_days):
            for i in range(int(d1*self._dose_interval), int(d2*self._dose_interval)):
                inflow = 1 - self._E_MAX * \
                    Cs_gamma[i] / (self._EC_50_gamma + Cs_gamma[i])
                self._A[0] = inflow
                self._A[7] = inflow
                self._A[1:] += self._ktr[1:] * (self._A[0:-1] - self._A[1:])

                # self._dA[0] = self._ktr1 * (1 - self._E_MAX * Cs_gamma[i] /
                #                             (self._EC_50 ** self._gamma + Cs_gamma[i])) - self._ktr1*self._A[0]
                # for j in range(1, 6):
                #     self._dA[j] = self._ktr1 * (self._A[j-1] - self._A[j])

                # self._dA[6] = self._ktr2 * (1 - self._E_MAX * Cs_gamma[i] /
                #                             (self._EC_50 ** self._gamma + Cs_gamma[i])) - self._ktr2*self._A[6]
                # for j in range(7):
                #     self._A[j] += self._dA[j]

            INR.append(
                (baseINR + (INR_max*(1-self._A[6]*self._A[8]) ** self._lambda)) * self._exp_e_INR[d2])
            # INR.append(
            #     (baseINR + (INR_max*(1-self._A[5]*self._A[6]) ** self._lambda)) * self._exp_e_INR[d2])

        self._last_computed_day = end_days[-1]

        return INR



# -*- coding: utf-8 -*-
'''
environment class
=================

This `environment` class provides a learning environment for any reinforcement
learning agent on any subject.


'''

import functools
import pathlib
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

from reil import agents as rlagents
from reil import stateful
from reil import subjects as rlsubjects
from reil import utils
from reil.datatypes import reildata
from reil.subjects.subject import Subject

AgentSubjectTuple = Tuple[str, str]


class Environment(stateful.Stateful):
    '''
    Provide a learning environment for agents and subjects.

    Attributes
    ----------

    Methods
    -------
        add: add a set of objects (agents/ subjects) to the environment.
        remove: remove objects (agents/ subjects) from the environment.
        assign: assign agents to subjects.
        elapse: move forward in time and interact agents and subjects.
        elapse_iterable: iterate over instances of each subject and interact agents in each subject instance.
        trajectory: extract (state, action, reward) trajectory.
        load: load an object (agent/ subject) or an environment.
        save: save an object (agent/ subject) or the current environment.

    Agents act on subjects and receive the reward of their action and the new state of subjects.
    Then agents learn based on this information to act better.
    '''

    def __init__(self,
                 agents: Optional[Dict[str, rlagents.Agent]] = None,
                 subjects: Optional[Dict[str, rlsubjects.Subject]] = None,
                 assignment_list: Optional[Sequence[AgentSubjectTuple]] = None,
                 episodes: int = 1,
                 max_steps: int = 10000,
                 termination: str = 'any',
                 reset: str = 'any',
                 learning_batch_size: int = -1,
                 learning_method: str = 'every step',
                 **kwargs: Any):
        '''
        Create a new environment.

        Arguments
        ---------
            name: the name of the environment
            episodes: number of episodes of run. (Default = 1)
            termination: when to terminate one episode. (Default = 'any')
                'any': terminate if any of the subjects is terminated.
                'all': terminate if all the subjects are terminated.
            reset: how to reset subjects after each episode. (Default = 'any')
                'any': reset only the subject that is terminated.
                'all': reset all subjects.
            learning_method: how to learn from each episode. (Default = 'every step')
                'every step': learn after every move.
                'history': learn after each episode.
        '''
        super().__init__(name=kwargs.get('name', __name__),
                         logger_name=kwargs.get('logger_name', __name__),
                         **kwargs)

        self._agents = {}
        self._subjects = {}
        self._assignment_list = {}
        self._episodes = episodes
        self._total_experienced_episodes = {}
        self._max_steps = max_steps
        self._termination = termination
        self._reset = reset
        self._learning_batch_size = learning_batch_size
        self._learning_method = learning_method

        if agents is not None:
            self.add(agents=agents)
        if subjects is not None:
            self.add(subjects=subjects)
        if assignment_list is not None:
            self.assign(assignment_list)


    def add(self,
            agents: Optional[Dict[str, rlagents.Agent]] = None,
            subjects: Optional[Dict[str, rlsubjects.Subject]] = None) -> None:
        '''
        Add agents or subjects to the environment.

        Arguments
        ---------
            agents: a dictionary consist of agent name and agent object. Names should be unique, otherwise overwritten.
            subjects: a dictionary consist of subject name and subject object. Names should be unique, otherwise overwritten.
        '''
        if agents is not None:
            self._agents.update(agents)

        if subjects is not None:
            self._subjects.update(subjects)
        # try:
        #     for name, agent in kwargs['agents'].items():
        #         self._agents[name] = agent
        # except (IndexError, KeyError):
        #     pass

        # try:
        #     for name, subject in kwargs['subjects'].items():
        #         self._subjects[name] = subject
        # except (IndexError, KeyError):
        #     pass

    def remove(self,
               agents: Optional[List[str]] = None,
               subjects: Optional[List[str]] = None) -> None:
        '''
        Remove agents or subjects from the environment.

        Arguments
        ---------
            agents: a list of agent names to be deleted.
            subjects: a list of subject names to be deleted.

        Raises KeyError if the agent is not found.
        '''
        if agents is not None:
            for name in agents:
                try:
                    del self._agents[name]
                except KeyError:
                    raise KeyError(f'Agent {name} not found!')

        if subjects is not None:
            for name in subjects:
                try:
                    del self._subjects[name]
                except KeyError:
                    raise KeyError(f'Subject {name} not found!')

    def assign(self, agent_subject_names: Sequence[AgentSubjectTuple]) -> None:
        '''
        Assign agents to subjects.

        Arguments
        ---------
            agent_subject_names: a list of agent subject tuples.

        Raises ValueError if an agent or subject is not found.
        Note: An agent cannot be assigned to act on multiple subjects, but a subject can be affected by multiple agents.
        '''
        for agent_name, subject_name in agent_subject_names:
            if agent_name not in self._agents:
                raise ValueError(f'Agent {agent_name} not found!')
            if subject_name not in self._subjects:
                raise ValueError(f'Subject {subject_name} not found!')
            _id = self._subjects[subject_name].register(agent_name)

            self._total_experienced_episodes[(agent_name, subject_name)] = 0
            self._assignment_list[(agent_name, subject_name)] = _id
            # try:
            #     self._assignment_list[agent_name].append((subject_name, _id))
            # except KeyError:
            #     self._assignment_list[agent_name] = [(subject_name, _id)]

    def divest(self, agent_subject_names: List[AgentSubjectTuple]) -> None:
        '''
        Divest agent subject assignment.

        Arguments
        ---------
            agent_subject_names: a list of agent subject tuples.

        Raises ValueError if an agent or subject is not found.
        Note: An agent can be assigned to act on multiple subjects and a subject can be affected by multiple agents.
        '''
        for agent_name, subject_name in agent_subject_names:
            if agent_name not in self._agents:
                raise ValueError(f'Agent {agent_name} not found!')
            if subject_name not in self._subjects:
                raise ValueError(f'Subject {subject_name} not found!')

            self._subjects[subject_name].deregister(agent_name)
            self._assignment_list.pop((agent_name, subject_name))
            del self._total_experienced_episodes[(agent_name, subject_name)]

    def elapse(self,
               episodes: Optional[int] = None,
               max_steps: Optional[int] = None,
               termination: Optional[str] = None,
               reset: Optional[str] = None,
               learning_method: Optional[str] = None,
               reporting: Optional[str] = None,
               tally: bool = False,
               step_count: bool = False):
        '''
        Move forward in time for a number of episodes.

        At each episode, agents are called sequentially to act on their respective subject.
        NOTE: This method loops over agents and only assumes one subject per agent.
        An episode ends if one (all) subject(s) terminates.
        Arguments
        ---------
            episodes: number of episodes of run.
            max_steps: maximum number of steps in each episode (Default = 10,000)
            termination: when to terminate one episode. (Default = 'any')
                'any': terminate if any of the subjects is terminated.
                'all': terminate if all the subjects are terminated.
            reset: how to reset subjects after each episode. (Default = 'any')
                'any': reset only the subject that is terminated.
                'all': reset all subjects.
            learning_method: how to learn from each episode. (Default = 'every step')
                'every step': learn after every move.
                'history': learn after each episode.
            reporting: what to report. (Default = 'none')
                'all': prints every move.
                'none' reports nothing.
                'important' reports important parts only.
            tally: count wins of each agent or not. (Default = 'no')
                'yes': counts the number of wins of each agent.
                'no': doesn't tally.
            step_count: count average number of steps in each episode. (Default = 'no')
                'yes': counts the average number of steps in each episode.
                'no': doesn't count.
        '''
        _episodes = episodes or self._episodes
        _max_steps = max_steps or self._max_steps
        if (termination or self._termination).lower() == 'all':
            def termination_func(x: bool, y: rlsubjects.Subject) -> bool:
                return x & y.is_terminated
            list_of_subjects: List[Subject] = list(self._subjects.values())
            reduce_initial_val = True

        else:
            def termination_func(x: bool, y: rlsubjects.Subject) -> bool:
                return x | y.is_terminated
            list_of_subjects: List[Subject] = list(self._subjects.values())
            reduce_initial_val = False

        _reset = (reset or self._reset).lower()
        _learning_method = (
            learning_method or self._learning_method).lower()
        _reporting = (reporting or 'none').lower()

        _tally = tally
        win_count: Dict[str, int] = dict((agent, 0) for agent in self._agents)

        _step_count = step_count

        # if _learning_method == 'none':
        #     for agent in self._agents.values():
        #         agent.training_mode = False
        # else:
        #     for agent in self._agents.values():
        #         agent.training_mode = True

        report_string = ''
        steps = 0
        for episode in range(_episodes):
            if _reporting != 'none':
                report_string = f'episode: {episode + 1}'
            # history = dict((agent_name, []) for agent_name in self._agents)
            history: Dict[str, List[Dict[str, reildata.ReilData]]] = defaultdict(list)
            done = False
            stopping_criterion = _max_steps * (episode + 1)
            while not done:
                if steps >= stopping_criterion:
                    break
                steps += 1
                for agent_name, agent in self._agents.items():
                    if not done:
                        subject_name, _id = self._assignment_list[agent_name]
                        subject = self._subjects[subject_name]
                        if not subject.is_terminated:
                            state = subject.state
                            possible_actions = subject.possible_actions
                            action = agent.act(state, actions=possible_actions,
                                               episode=self._total_experienced_episodes[(agent_name, subject_name)])
                            reward = subject.take_effect(action, _id)

                            if _reporting == 'all':
                                print(
                                    f'step: {steps: 4} episode: {episode:2} state: {state} action: {action} by:{agent_name}')

                            history[agent_name].append(
                                {'state': state, 'action': action, 'reward': reward})

                            if subject.is_terminated:
                                win_count[agent_name] += int(reward[0].value > 0)
                                for affected_agent in self._agents:
                                    if (self._assignment_list[affected_agent][0] == subject_name) & \
                                            (affected_agent != agent_name):
                                        history[affected_agent][-1]['reward'] = -reward

                    done = functools.reduce(termination_func, list_of_subjects, reduce_initial_val)

                if _learning_method == 'every step':
                    raise NotImplementedError
                    # for agent_name, agent in self._agents.items():
                    #     agent.learn(state=self._subjects[self._assignment_list[agent_name][0]].state,
                    #                 reward=history[agent_name][-1]['reward'])
                    #     history[agent_name] = []

            if _learning_method == 'history':
                for agent_name, agent in self._agents.items():
                    agent.learn(history=history[agent_name])

            for agent_name, subject_name in self._total_experienced_episodes:
                if self._subjects[subject_name].is_terminated:
                    self._total_experienced_episodes[(
                        agent_name, subject_name)] += 1

            if _reset == 'all':
                for agent in self._agents.values():
                    agent.reset()
                for subject in self._subjects.values():
                    subject.reset()
            elif _reset == 'any':
                for agent_name, subject_name in self._assignment_list.items():
                    if self._subjects[subject_name[0]].is_terminated:
                        self._agents[agent_name].reset()
                        self._subjects[subject_name[0]].reset()

            if _tally & (_reporting != 'none'):
                report_string += f'\n tally:'
                for agent in self._agents:
                    report_string += f'\n {agent} {win_count[agent]}'

            if _reporting != 'none':
                print(report_string)

        if _tally:
            return win_count
        if _step_count:
            return steps/_episodes

    def elapse_iterable(self,
                        max_steps: Optional[int] = None,
                        test_mode: Sequence[AgentSubjectTuple] = (),
                        reset: str = 'any',
                        return_output: Sequence[AgentSubjectTuple] = (),
                        stats: Optional[Dict[AgentSubjectTuple, Sequence[str]]] = None,
                        stats_func: Optional[Dict[AgentSubjectTuple,
                                                  Callable[[Any], Any]]] = None
                        ) -> Tuple[Dict[AgentSubjectTuple, List[Any]], Dict[AgentSubjectTuple, List[Any]]]:
        '''
        Move forward in time until IterableSubject is consumed.

        For each IterableSubject, agents are called sequentially to act based on the assignment list.
        Method ends if all IterableSubjects are consumed.
        Arguments
        ---------
            max_steps: maximum number of steps in each episode (Default = 10,000)
            training_mode: whether it is in training or test mode. (Default: True)
            reset: whether to reset `agents`. 'any` resets agents who acted on a finished subject. `all` resets all agents. (Default = 'any')
            return_output: a dictionary that indicates whether to return the resulting outputs or not (Default: False)
            stats: a dictionary that contains stats for each agent-subject pair
            stats_func: a dictionary that contains functions to calculate stats.
        '''

        _max_steps = max_steps or self._max_steps
        _training_mode: Dict[AgentSubjectTuple, bool] = defaultdict(lambda: True)
        for a_s_tuple in test_mode:
            _training_mode[a_s_tuple] = False

        if reset.lower() not in ('any', 'all'):
            self._logger.warning(f'reset argument should be either "any" or "all". Received {reset}. "any" is assumed.')
            _reset = 'any'
        else:
            _reset = reset.lower()

        _return_output = defaultdict(bool)
        for a_s_tuple in return_output:
            _return_output[a_s_tuple] = True

        _stats = stats or defaultdict(list)

        if stats_func is None and stats is not None:
            self._logger.exception('stats are provided, but no stats_func is provided.')
            raise ValueError('stats are provided, but no stats_func is provided.')
        temp = stats_func or lambda _: None
        if isinstance(temp, Callable):
            _stats_func = dict((agent_subject_tuple, temp) for agent_subject_tuple in self._assignment_list)
        else:
            _stats_func = temp

        output = defaultdict(list)
        stats_agents = defaultdict(list)
        stats_subjects = defaultdict(list)
        stats_final = defaultdict(list)

        for subject_name, subject in self._subjects.items():
            assigned_agents = list((k[0], v)
                for k, v in self._assignment_list.items()
                if k[1] == subject_name)

            if assigned_agents == []:
                continue

            for agent_name, _ in assigned_agents:
                self._agents[agent_name].training_mode = _training_mode[(
                    agent_name, subject_name)]

            history: Dict[str, List[Dict[str, reildata.ReilData]]] = defaultdict(list)
            for instance_id, subject_instance in subject:
                steps = 0
                # for agent_name, _ in assigned_agents:
                #     self._agents[agent_name].exchange_protocol = subject_instance.requested_exchange_protocol
                while not subject_instance.is_terminated():
                    for agent_name, _id in assigned_agents:
                        agent = self._agents[agent_name]
                        if subject_instance.is_terminated() or steps >= _max_steps:
                            break
                        steps += 1

                        complete_state = None
                        try:
                            complete_state = subject_instance.complete_state
                        except NotImplementedError:
                            pass

                        state = subject_instance.state()
                        possible_actions = subject_instance.possible_actions()
                        action = agent.act(state, actions=possible_actions,
                                           episode=self._total_experienced_episodes[(agent_name, subject_name)])
                        subject_instance.take_effect(action, _id)
                        reward = subject_instance.reward()

                        history[agent_name].append({'instance_id': instance_id,
                                                    'state': state,
                                                    'action': action,
                                                    'reward': reward,
                                                    'complete_state': complete_state})

                        if subject_instance.is_terminated():
                            for affected_agent in self._agents:
                                if (affected_agent, subject_name) in self._assignment_list and \
                                        (affected_agent != agent_name):
                                    history[affected_agent][-1]['reward'] = -reward

                            complete_state = None
                            if subject_instance.exchange_protocol['complete_state']:
                                try:
                                    complete_state = subject_instance.complete_state
                                except NotImplementedError:
                                    pass
                            history[agent_name].append({'instance_id': instance_id,
                                                        'state': subject_instance.state,
                                                        'complete_state': complete_state})

                        if _training_mode[(agent_name, subject_name)] \
                                and self._learning_batch_size != -1 \
                                and len(history[agent_name]) >= self._learning_batch_size:
                            agent.learn(history=history[agent_name])
                            history[agent_name] = []

                for agent_name, _ in assigned_agents:
                    agent_subject_tuple = (agent_name, subject_name)
                    if _training_mode[agent_subject_tuple]:
                        self._total_experienced_episodes[agent_subject_tuple] += 1

                        if self._learning_batch_size == -1:
                            self._agents[agent_name].learn(
                                history=history[agent_name])

                    if _return_output[agent_subject_tuple]:
                        output[agent_subject_tuple].append(history[agent_name])

                    history[agent_name] = []

                    try:
                        stats_list = stats[agent_subject_tuple]
                        result_agent = self._agents[agent_name].stats(
                            stats_list)
                        result_subject = subject_instance.stats(stats_list)

                        if result_agent:
                            stats_agents[agent_subject_tuple].append(
                                result_agent)

                        if result_subject:
                            stats_subjects[agent_subject_tuple].append(
                                result_subject)
                    except KeyError:
                        pass

            if _reset == 'all':
                for agent in self._agents.values():
                    agent.reset()
            elif _reset == 'any':
                for agent_name, _ in assigned_agents:
                    self._agents[agent_name].reset()

            for agent_name, _ in assigned_agents:
                agent_subject_tuple = (agent_name, subject_name)

                if _stats.get(agent_subject_tuple, []):
                    result = _stats_func[agent_subject_tuple](agent_stats=stats_agents.get(agent_subject_tuple, None),
                                                                subject_stats=stats_subjects.get(agent_subject_tuple, None))
                    stats_final[agent_subject_tuple].append(result)

        return stats_final, output

    # TODO: Make arguments explicit + type annotation
    def trajectory(self, **kwargs: Any):
        '''
        Extract (state, action, reward) trajectory.

        At each episode, agents are called sequentially to act on their respective subject(s).
        An episode ends if one (all) subject(s) terminates.
        Arguments
        ---------
            max_steps: maximum number of steps in the episode (Default = 10,000)
            termination: when to terminate one episode. (Default = 'any')
                'any': terminate if any of the subjects is terminated.
                'all': terminate if all the subjects are terminated.
        '''
        max_steps = kwargs.get('max_steps', self._max_steps)

        if kwargs.get('termination', self._termination).lower() == 'all':
            def termination_func(x: bool, y: rlsubjects.Subject) -> bool:
                return x & y.is_terminated
            list_of_subjects = list(self._subjects.values())
            reduce_initial_val = True
        else:
            def termination_func(x: bool, y: rlsubjects.Subject) -> bool:
                return x | y.is_terminated
            list_of_subjects = list(self._subjects.values())
            reduce_initial_val = False

        for agent in self._agents.values():
            agent.training_mode = False

        for subject in self._subjects.values():
            subject.reset()

        history = dict((agent_subject, [])
                       for agent_subject in self._assignment_list)

        done = False
        steps = 0
        while not done:
            if steps >= max_steps:
                break
            steps += 1
            for agent_name, subject_name in self._assignment_list:
                if not done:
                    _id = self._assignment_list[(agent_name, subject_name)]
                    agent = self._agents[agent_name]
                    subject = self._subjects[subject_name]
                    if not subject.is_terminated:
                        state = subject.state
                        possible_actions = subject.possible_actions
                        action = agent.act(state, actions=possible_actions)
                        q = float(agent._q(state, action))
                        reward = subject.take_effect(action, _id)

                        history[(agent_name, subject_name)].append(
                            {'state': state, 'action': action, 'q': q, 'reward': reward})
                        if subject.is_terminated:
                            for affected_agent in self._agents.keys():
                                if ((affected_agent, subject) in self._assignment_list.keys()) & \
                                        (affected_agent != agent_name):
                                    history[(affected_agent, subject)
                                            ][-1]['reward'] = -reward

                done = functools.reduce(termination_func, list_of_subjects, reduce_initial_val)

        for sub in self._subjects.values():
            sub.reset()

        return history

    def load(self,
             object_name: Union[List[str], str] = 'all',
             filename: Optional[str] = None,
             path: Optional[Union[pathlib.Path, str]] = None) -> None:
        '''
        Load an object or an environment from a file.
        Arguments
        ---------
            filename: the name of the file to be loaded.
            object_name: if specified, that object (agent or subject) is being loaded from file. 'all' loads an environment. (Default = 'all')
        Raises ValueError if the filename is not specified.
        '''
        _filename: str = filename or self._name
        _path = pathlib.Path(path if path is not None else self._path)

        if object_name == 'all':
            super().load(filename=_filename, path=_path)
            self._agents: Dict[str, rlagents.Agent] = {}
            self._subjects: Dict[str, rlsubjects.Subject] = {}
            for name, obj_type in self._env_data['agents']:  # type: ignore
                self._agents[name] = obj_type.from_pickle(  # type: ignore
                    path=(_path / f'{_filename}.data'), filename=name)
            for name, obj_type in self._env_data['subjects']:  # type: ignore
                self._subjects[name] = obj_type.from_pickle(  # type: ignore
                    path=(_path / f'{_filename}.data'), filename=name)

            del self._env_data  # type: ignore

        else:
            for obj in object_name:
                if obj in self._agents:
                    self._agents[obj].load(
                        path=(_path / f'{_filename}.data'), filename=obj)
                    self._agents[obj].reset()
                elif obj in self._subjects:
                    self._subjects[obj].load(
                        path=(_path / f'{_filename}.data'), filename=obj)
                    self._subjects[obj].reset()

    def save(self,
             filename: Optional[str] = None,
             path: Optional[Union[pathlib.Path, str]] = None,
             data_to_save: Union[List[str], str] = 'all') -> Tuple[pathlib.Path, str]:
        '''
        Save an object or the environment to a file.
        Arguments
        ---------
            filename: the name of the file to be saved.
            path: the path of the file to be saved. (Default='./')
            object_name: if specified, that object (agent or subject) is being saved to file. 'all' saves the environment. (Default = 'all')
        Raises ValueError if the filename is not specified.
        '''
        _filename = filename or self._name
        _path = pathlib.Path(path if path is not None else self._path)

        if data_to_save == 'all':
            self._env_data: Dict[str, List[Any]] = defaultdict(list)

            for name, agent in self._agents.items():
                _, fn = agent.save(
                    path=_path / f'{_filename}.data', filename=name)
                self._env_data['agents'].append((fn, type(agent)))

            for name, subject in self._subjects.items():
                _, fn = subject.save(
                    path=_path / f'{_filename}.data', filename=name)
                self._env_data['subjects'].append((fn, type(subject)))

            super().save(filename=_filename, path=_path,
                         data_to_save=[v for v in self.__dict__
                            if v not in ('_agents', '_subjects')])

            del self._env_data
        else:
            for obj in data_to_save:
                if obj in self._agents:
                    self._agents[obj].save(
                        path=_path / f'{_filename}.data', filename=obj)
                elif obj in self._subjects:
                    self._subjects[obj].save(
                        path=_path / f'{_filename}.data', filename=obj)

        return _path, _filename

    def __repr__(self) -> str:
        try:
            return 'Env: \n Agents:\n' + \
                '\n\t'.join((a.__repr__() for a in self._agents.values())) + \
                '\nSubjects:\n' + \
                '\n\t'.join((s.__repr__() for s in self._subjects.values()))
        except AttributeError:
            return 'Environment: New'



# -*- coding: utf-8 -*-
'''
WarfarinPatientRavvaz class
===========================

A warfarin patient class with features and parameters of Ravvaz et al. 2016.

Features included in this model are:
* age
* weight
* height
* gender
* race
* tobaco
* amiodarone
* fluvastatin
* CYP2C9
* VKORC1
* MTT_1
* MTT_2
* cyp_1_1
* V1
* V2
* EC_50
'''
import math
from typing import Any

from reil.datatypes import Feature
from reil.subjects.healthcare import Patient
from reil.subjects.healthcare.mathematical_models import HealthMathModel
from reil.utils.functions import (random_categorical, random_truncated_lnorm,
                                  random_truncated_normal)


class WarfarinPatientRavvaz(Patient):
    def __init__(self,
                 model: HealthMathModel,
                 randomized: bool = True,
                 **feature_values: Any) -> None:
        '''
        Parameters
        ----------
        model:
            A `HealthMathModel` to be used to model patient's behavior.

        randomized:
            Whether patient characteristics and model parameters should be
            generated randomly or deterministically.

        feature_values:
            Keyword arguments by which some of the `features` of the patient
            can be determined. For example, if "age" is one of the features,
            age=40.0 will set the initial age to 40.0.
        '''
        self.feature_set = {
            'age': Feature.numerical(  # Aurora population
                lower=18.0, upper=100.0, mean=67.30, stdev=13.43,
                generator=random_truncated_normal),
            'weight': Feature.numerical(  # lb  - Aurora population
                lower=70.0, upper=500.0, mean=199.24, stdev=54.71,
                generator=random_truncated_normal),
            'height': Feature.numerical(  # in - Aurora population
                lower=45.0, upper=85.0, mean=66.78, stdev=4.31,
                generator=random_truncated_normal),
            'gender': Feature.categorical(  # Aurora population
                categories=('Female', 'Male'),
                probabilities=(0.5314, 0.4686),
                generator=random_categorical),
            'race': Feature.categorical(  # Aurora Avatar Population
                categories=('White', 'Black', 'Asian',
                            'American Indian', 'Pacific Islander'),
                probabilities=(0.9522, 0.0419, 0.0040, 0.0018, 1e-4),
                generator=random_categorical),
            'tobaco': Feature.categorical(  # Aurora Avatar Population
                categories=('No', 'Yes'),
                probabilities=(0.9067, 0.0933),
                generator=random_categorical),
            'amiodarone': Feature.categorical(  # Aurora Avatar Population
                categories=('No', 'Yes'),
                probabilities=(0.8849, 0.1151),
                generator=random_categorical),
            'fluvastatin': Feature.categorical(  # Aurora Avatar Population
                categories=('No', 'Yes'),
                probabilities=(0.9998, 0.0002),
                generator=random_categorical),
            'CYP2C9': Feature.categorical(  # Aurora Avatar Population
                categories=('*1/*1', '*1/*2', '*1/*3',
                            '*2/*2', '*2/*3', '*3/*3'),
                probabilities=(0.6739, 0.1486, 0.0925, 0.0651, 0.0197, 2e-4),
                generator=random_categorical),
            'VKORC1': Feature.categorical(  # Aurora Avatar Population
                categories=('G/G', 'G/A', 'A/A'),
                probabilities=(0.3837, 0.4418, 0.1745),
                generator=random_categorical),

            'MTT_1': Feature.numerical(  # Hamberg PK/PD
                value=math.exp(math.log(11.6) + 0.141/2), mean=math.log(11.6),
                stdev=math.sqrt(0.141),
                generator=random_truncated_lnorm),
            'MTT_2': Feature.numerical(  # Hamberg PK/PD
                value=math.exp(math.log(120.0) + 1.02/2), mean=math.log(120.0),
                stdev=math.sqrt(1.02),
                generator=random_truncated_lnorm),
            'cyp_1_1': Feature.numerical(  # Hamberg PK/PD
                value=math.exp(math.log(0.314) + 0.31/2), mean=math.log(0.314),
                stdev=math.sqrt(0.31),
                generator=random_truncated_lnorm),
            'V1': Feature.numerical(  # Hamberg PK/PD
                value=math.exp(math.log(13.8) + 0.262/2), mean=math.log(13.8),
                stdev=math.sqrt(0.262),
                generator=random_truncated_lnorm),
            'V2': Feature.numerical(  # Hamberg PK/PD
                value=math.exp(math.log(6.59) + 0.991/2), mean=math.log(6.59),
                stdev=math.sqrt(0.991),
                generator=random_truncated_lnorm),

            'EC_50': Feature.numerical(  # Hamberg PK/PD
                stdev=math.sqrt(0.409),
                generator=random_truncated_lnorm),

            'sensitivity': Feature.categorical(
                categories=('normal', 'sensitive', 'highly sensitive'),
                generator=lambda f: f.value)
        }

        for f in self.feature_set.values():
            f.randomized = randomized

        # Since EC_50 is not set (it depends on other features),
        # super().__init__() fails to setup the model.
        # I catch it, generate EC_50 and set up the model.
        try:
            super().__init__(model, EC_50=None, **feature_values)
        except TypeError:
            self._generate_EC_50()
            self._generate_sensitivity()
            self._model.setup(**self.feature_set)

    def generate(self) -> None:
        super().generate()
        self._generate_EC_50()
        self._generate_sensitivity()
        self._model.setup(**self.feature_set)

    def _generate_EC_50(self) -> None:
        if self.feature_set['VKORC1'].value == 'G/G':
            self.feature_set['EC_50'].value = math.log(4.61) + 0.409/2
            self.feature_set['EC_50'].mean = math.log(4.61)
        elif self.feature_set['VKORC1'].value in ['G/A', 'A/G']:
            self.feature_set['EC_50'].value = math.log(3.02) + 0.409/2
            self.feature_set['EC_50'].mean = math.log(3.02)
        elif self.feature_set['VKORC1'].value == 'A/A':
            self.feature_set['EC_50'].value = math.log(2.20) + 0.409/2
            self.feature_set['EC_50'].mean = math.log(2.20)

        self.feature_set['EC_50'].generate()

    def _generate_sensitivity(self):
        combo = (self.feature_set['CYP2C9'].value +
                 self.feature_set['VKORC1'].value)

        if combo in ('*1/*1G/G', '*1/*2G/G', '*1/*1G/A'):
            s = 'normal'
        elif combo in ('*1/*2G/A', '*1/*3G/A', '*2/*2G/A',
                       '*2/*3G/G', '*1/*3G/G', '*2/*2G/G',
                       '*1/*2A/A', '*1/*1A/A'):
            s = 'sensitive'
        elif combo in ('*3/*3G/G',
                       '*3/*3G/A', '*2/*3G/A',
                       '*3/*3A/A', '*2/*3A/A', '*2/*2A/A', '*1/*3A/A'):
            s = 'highly sensitive'
        else:
            raise ValueError(
                f'Unknown CYP2C9 and VKORC1 combination: {combo}.')

        self.feature_set['sensitivity'].value = s



# -*- coding: utf-8 -*-
'''
subject class
=============

This `subject` class is the base class of all subject classes.
'''

from typing import Any, Optional, Tuple, TypeVar

from reil import stateful
from reil.datatypes import ReilData, SecondayComponent


class Subject(stateful.Stateful):
    '''
    The base class of all subject classes.
    '''

    def __init__(self,
                 sequential_interaction: bool = True,
                 **kwargs: Any):
        '''
        Arguments
        ---------
        sequential_interaction:
            If `True`, `agents` can only act on the `subject` in the order they
            are added.

        Notes
        -----
        `sequential_interaction` is not enforced (implemented) yet!
        '''
        super().__init__(**kwargs)

        self._sequential_interaction = sequential_interaction
        self.reward = SecondayComponent(name='reward',
                                        primary_component=self.state,
                                        enabled=False)

    def _default_reward_definition(
            self, _id: Optional[int] = None) -> ReilData:
        return ReilData.single_base(name='default_reward', value=None)

    def is_terminated(self, _id: Optional[int] = None) -> bool:
        '''
        Determine if the `subject` is terminated for the given `agent` ID.

        Arguments
        ---------
        _id:
            ID of the agent that checks termination. In a multi-agent setting,
            e.g. an RTS game, one agent might die and another agent might still
            be alive.

        Returns
        -------
        :
            `False` as long as the subject can accept new actions from the
            `agent`. If `_id` is `None`, then returns `True` if no `agent`
            can act on the `subject`.
        '''
        raise NotImplementedError

    def possible_actions(self, _id: int = 0) -> Tuple[ReilData, ...]:
        '''
        Generate the list of possible actions.

        Arguments
        ---------
        _id:
            ID of the `agent` that wants to act on the `subject`.

        Returns
        -------
        :
            A list of possible actions for the `agent` with ID=_id.
        '''
        return (ReilData.single_base(name='default_action'),)

    def take_effect(self, action: ReilData, _id: int = 0) -> None:
        '''
        Receive an `action` from `agent` with ID=`_id` and transition to
        the next state.

        Arguments
        ---------
        action:
            The action sent by the `agent` that will affect this `subject`.

        _id:
            ID of the `agent` that has sent the `action`.
        '''
        self.reward.enable()

    def reset(self) -> None:
        '''Reset the `subject`, so that it can resume accepting actions.'''
        self.reward.disable()


SubjectType = TypeVar('SubjectType', bound=Subject)

# def reward(self,
#            name: str = 'default', _id: int = 0) -> reildata.ReilData:
#     '''
#     Compute the reward that `agent` receives, based on the reward
#     definition `name`.

#     Arguments
#     ---------
#     name:
#         The name of the reward definition. If omitted, output of the
#         `default_reward` method will be returned.

#     _id:
#         The ID of the calling `agent`.

#     Returns
#     -------
#     :
#         The reward for the given `agent`.
#     '''
#     return self._reward(name, _id)


# def reward(self,
#            _id: int = 0, name: Optional[str] = None) -> reildata.ReilData:
#     '''
#     Compute the reward that `agent` receives, based on the reward
#     definition `name`.

#     Arguments
#     ---------
#     _id:
#         The ID of the calling `agent`.

#     name:
#         The name of the reward definition. If omitted, output of the
#         `default_reward` method will be returned.

#     Returns
#     -------
#     :
#         The reward for the given `agent`.
#     '''
#     if name is None or name.lower() == 'default':
#         return self.default_reward(_id)

#     f, s = self._reward_definitions[name.lower()]
#     temp = f(self.state(s, _id))

#     return reildata.ReilData.single_base(name='reward', value=temp)

# def default_reward(self, _id: int = 0) -> reildata.ReilData:
#     '''
#     Compute the default reward definition of the subject for agent `_id`.

#     Arguments
#     ---------
#     _id:
#         ID of the `agent` that calls the reward method.

#     Returns
#     -------
#     :
#         The reward for the given `agent`.
#     '''
#     return reildata.ReilData.single_base(name='reward', value=0.0)

# def add_reward_definition(self, name: str,
#                           rl_function: reil_functions.ReilFunction,
#                           state_name: str) -> None:
#     '''
#     Add a new reward definition called `name` with function `rl_function`
#     that uses state `state_name`.

#     Arguments
#     ---------
#     name:
#         The name of the new reward definition.

#     rl_function:
#         An instance of `ReilFunction` that gets the state of the
#         `subject`, and computes the reward. The `rl_function` should
#         have the list of arguments from the state in its definition.

#     state_name:
#         The name of the state definition that should be used to
#         compute the reward.

#     Raises
#     ------
#     ValueError:
#         The reward `name` already exists.

#     ValueError:
#         The `state_name` is undefined.

#     Notes
#     -----
#         `statistic` and `reward` are basicly doing the same thing. The
#         difference is in their application: `statistic` should be called at
#         the end of each trajectory (sampled path) to compute the necessary
#         statistics about the performance of the `agents` and `subjects`.
#         `reward`, on the other hand, should be called after each
#         interaction between an `agent` and the `subject` to guide the
#         reinforcement learning model to learn the optimal policy.
#     '''
#     if name.lower() in self._reward_definitions:
#         raise ValueError(f'Reward definition {name} already exists.')

#     if state_name.lower() not in self._state_definitions:
#         raise ValueError(f'Unknown state name: {state_name}.')

#     self._reward_definitions[name.lower()] = (rl_function, state_name)



# -*- coding: utf-8 -*-
'''
agents module for reinforcement learning
========================================

This module provides different agents in reinforcement learning context.

Classes
-------
NoLearnAgent
    the base class of all agent classes

Agent
    the base class of all agent classes that learn from history

QLearning
    the Q-learning agent that can accept any learner

DeepQLearning
    the agent with neural network as a learner (derived from
    `QLearning` class)

RandomAgent
    an agent that randomly chooses an action

UserAgent
    an agent that shows current state and asks for user's choice
    of action

WarfarinAgent
    an agent based on Ravvaz et al (2016) paper for Warfarin Dosing

Types
-----
TrainingData
    a type alias for training data, consisting of a tuple of
    `ReilData` for X matrix and a tuple of floats for Y vector.

AgentType
    a type variable that is bound to `Agent`.
'''
from .no_learn_agent import NoLearnAgent, TrainingData  # noqa: W0611
from .agent import Agent  # noqa: W0611
from .q_learning import QLearning  # noqa: W0611
from .deep_q_learning import DeepQLearning  # noqa: W0611
from .random_agent import RandomAgent  # noqa: W0611
from .user_agent import UserAgent  # noqa: W0611
from .warfarin_agent import WarfarinAgent  # noqa: W0611

from typing import TypeVar

AgentType = TypeVar('AgentType', bound=Agent)



# -*- coding: utf-8 -*-
'''
WarfarinAgent class
===================

An agent for warfarin modeling based on the doses define in Ravvaz et al (2017)
'''

import collections
import functools
from math import exp, log, sqrt
from typing import Any, Dict, List, Optional, Tuple

from reil import agents
from reil.datatypes.reildata import ReilData
from typing_extensions import Literal

DoseInterval = collections.namedtuple('DoseInterval', 'dose, interval')


class DosingProtocol:
    '''
    Base class for all dosing protocol objects.
    '''

    def __init__(self) -> None:
        pass

    def prescribe(self,
                  patient: Dict[str, Any],
                  additional_info: Dict[str, Any]
                  ) -> Tuple[float, int, Dict[str, Any]]:
        '''
        Prescribe a dose for the given `patient` and `additional_info`.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics necessary to make dosing
            decisions.

        additional_info:
            A dictionary of information being communicated between protocols at
            each call to `prescribe`. These additional information are
            protocol-dependent.

        Returns
        -------
        :
            The prescribed dose along with updated `additional_info`.
        '''
        raise NotImplementedError

    def reset(self) -> None:
        '''Reset the dosing protocol'''
        pass


class Aurora(DosingProtocol):
    '''
    Aurora Dosing Protocol, based on Ravvaz et al. (2017)
    '''

    def prescribe(self,  # noqa: C901
                  patient: Dict[str, Any],
                  additional_info: Dict[str, Any]
                  ) -> Tuple[float, int, Dict[str, Any]]:
        today = patient['day']
        INRs = patient['INRs']
        previous_INR = INRs[-1]
        previous_dose = patient['Doses'][-1]
        previous_interval = patient['Intervals'][-1]

        red_flag = additional_info.get('red_flag', False)
        skip_dose = additional_info.get('skip_dose', 0)
        new_dose = additional_info.get('new_dose', 0)
        number_of_stable_days = additional_info.get('number_of_stable_days', 0)

        if red_flag:
            if previous_INR > 3.0:
                next_dose = 0.0
                next_interval = 2
            else:
                red_flag = False
                next_dose = new_dose
                next_interval = 7
        elif skip_dose:
            skip_dose = 0
            next_dose = new_dose
            next_interval = 7
        elif today <= 2:  # initial dosing
            next_dose = 10.0 if patient['age'] < 65.0 else 5.0
            next_interval = 3 - today
        elif today == 3:  # adjustment dosing
            if previous_INR >= 2.0:
                next_dose = 5.0
                next_interval = 2
                if previous_INR <= 3.0:
                    number_of_stable_days = self._stable_days(
                        INRs[-2], previous_INR, previous_interval)
            else:
                next_dose, next_interval, _, _ = self.aurora_dosing_table(
                    previous_INR, previous_dose)
        elif today == 4:
            raise ValueError(
                'Cannot use Aurora on day 4. '
                'Dose on day 4 equals dose on day 3.')
        else:  # maintenance dosing
            if 2 <= previous_INR <= 3:
                number_of_stable_days += self._stable_days(
                    INRs[-2], previous_INR, previous_interval)

                next_dose = previous_dose
                next_interval = self.aurora_retesting_table(
                    number_of_stable_days)
            else:
                number_of_stable_days = 0
                new_dose, next_interval, skip_dose, red_flag = \
                    self.aurora_dosing_table(previous_INR, previous_dose)
                if red_flag:
                    next_dose = 0.0
                    next_interval = 2
                elif skip_dose:
                    next_dose = 0.0
                    next_interval = skip_dose
                else:
                    next_dose = new_dose

        new_info = {
            'red_flag': red_flag,
            'skip_dose': skip_dose,
            'new_dose': new_dose,
            'number_of_stable_days': number_of_stable_days
        }

        return next_dose, next_interval, new_info

    @staticmethod
    def _stable_days(INR_start: float, INR_end: float, interval: int) -> int:
        '''
        Interpolate the INR values in the range and compute the number of days
        in therapeutic range of [2, 3].

        Arguments
        ---------
        INR_start:
            INR at the beginning of the period.

        INR_end:
            INR at the end of the period.

        interval:
            The number of days from start to end.

        Returns
        -------
        :
            The number of days in therapeuric range (TTR).

        Notes
        -----
        This method excludes `INR_start` and includes `INR_end` in
        computing TTR.

        '''
        return sum(2 <= INR_end + (INR_start - INR_end) * i / interval <= 3
                   for i in range(1, interval+1))

    @staticmethod
    def aurora_dosing_table(
            current_INR: float, dose: float) -> Tuple[float, int, int, bool]:
        '''
        Determine the dosing information, based on Aurora dosing table.

        Arguments
        ---------
        current_INR:
            The latest value of INR.

        dose:
            The latest dose prescribed.

        Returns
        -------
        :
            * The next dose
            * The time of the next test (in days).
            * The number of doses to skip.
            * Red flag for too high INR values.
        '''
        skip_dose = 0
        red_flag = False
        if current_INR < 1.50:
            dose = dose * 1.15
            next_test = 7
        elif current_INR < 1.80:
            dose = dose * 1.10
            next_test = 7
        elif current_INR < 2.00:
            dose = dose * 1.075
            next_test = 7
        elif current_INR <= 3.00:
            next_test = 28
        elif current_INR < 3.40:
            dose = dose * 0.925
            next_test = 7
        elif current_INR < 4.00:
            dose = dose * 0.9
            next_test = 7
        elif current_INR <= 5.00:
            skip_dose = 2
            dose = dose * 0.875
            next_test = 7
        else:
            red_flag = True
            next_test = 2
            dose = dose * 0.85

        return dose, next_test, skip_dose, red_flag

    @staticmethod
    def aurora_retesting_table(number_of_stable_days: int) -> int:
        '''
        Determine when the next test should be based on current number of
        stable days.

        Arguments
        ---------
        number_of_stable_days:
            Number of consecutive stable days.

        Returns
        -------
        :
            The time of the next test (in days).
        '''
        retesting_table = {1: 1, 2: 5, 7: 7, 14: 14, 28: 28}
        return retesting_table[max(
            map(lambda x: x if x <= number_of_stable_days else 1,
                retesting_table))]


class IWPC(DosingProtocol):
    '''
    IWPC dosing protocol ('pharmacogenetic', 'clinical', 'modified').
    '''

    def __init__(self,
                 method: Literal['pharmacogenetic',
                                 'clinical',
                                 'modified'] = 'pharmacogenetic') -> None:
        '''
        Arguments
        ---------
        method:
            One of 'pharmacogenetic', 'clinical', 'modified'.
        '''
        if method.lower() == 'clinical':
            self._method = self.clinical
        elif method.lower() == 'modified':
            self._method = self.modified_pg
        elif method.lower() in ['pg', 'pharmacogenetic', 'default']:
            self._method = self.pg

        self.reset()

    def prescribe(self,
                  patient: Dict[str, Any],
                  additional_info: Dict[str, Any]
                  ) -> Tuple[float, int, Dict[str, Any]]:
        if self._doses:
            dose = self._doses.pop()
            interval = 1
        else:
            dose, interval = self._method(patient)
            if isinstance(dose, list):
                self._doses = dose
                self._doses.reverse()
                dose = self._doses.pop()
                interval = 1

        return dose, interval, {}

    @staticmethod
    def clinical(patient: Dict[str, Any]) -> Tuple[float, int]:
        '''
        Determine warfarin dose using clinical IWPC formula.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - height (in)
            - weight (lb)
            - race ('Asian', 'Black', etc.)
            - amiodarone ('yes', 'no')

        Returns
        -------
        :
            The dose and time for the next test.

        Notes
        -----
        This method always returns 2 (days) for the next test.
        '''
        weekly_dose = (4.0376
                       - 0.2546 * (patient['age'] // 10)
                       # in to cm
                       + 0.0118 * patient['height'] * 2.54
                       # lb to kg
                       + 0.0134 * patient['weight'] * 0.454
                       - 0.6752 * (patient['race'] == 'Asian')
                       + 0.4060 * (patient['race'] == 'Black')
                       # # missing or mixed race
                       # + 0.0443 * (patient['race'] not in  [...])
                       # Enzyme inducer status
                       # (Fluvastatin is reductant not an inducer!)
                       + 1.2799 * 0
                       - 0.5695 * (patient['amiodarone'] == 'Yes')) ** 2

        return weekly_dose / 7.0, 2

    # only the initial dose (day <= 2)
    @staticmethod
    def pg(patient: Dict[str, Any]) -> Tuple[float, int]:
        '''
        Determine warfarin dose using pharmacogenetic IWPC formula.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - height (in)
            - weight (lb)
            - VKORC1 ('G/G', 'G/A', 'A/A', etc.)
            - CYP2C9 ('*1/*1', '*1/*2', '*1/*3', '*2/*2', '*2/*3', '*3/*3',...)
            - race ('Asian', 'Black', etc.)
            - amiodarone ('yes', 'no')

        Returns
        -------
        :
            The dose and time for the next test.

        Notes
        -----
        This method always returns 2 (days) for the next test.
        '''
        weekly_dose = (5.6044
                       # Based on Ravvaz (EU-PACT (page 18) uses year, Ravvaz
                       # (page 10 of annex) uses decades!
                       - 0.2614 * (patient['age'] // 10)
                       + 0.0087 * patient['height'] * 2.54  # in to cm
                       + 0.0128 * patient['weight'] * 0.454  # lb to kg
                       - 0.8677 * (patient['VKORC1'] == 'G/A')
                       - 1.6974 * (patient['VKORC1'] == 'A/A')
                       # Not in EU-PACT ?!!
                       - 0.4854 * \
                       int(patient['VKORC1'] not in [
                           'G/A', 'A/A', 'G/G'])
                       - 0.5211 * (patient['CYP2C9'] == '*1/*2')
                       - 0.9357 * (patient['CYP2C9'] == '*1/*3')
                       - 1.0616 * (patient['CYP2C9'] == '*2/*2')
                       - 1.9206 * (patient['CYP2C9'] == '*2/*3')
                       - 2.3312 * (patient['CYP2C9'] == '*3/*3')
                       # Not in EU-PACT
                       - 0.2188 * \
                       int(patient['CYP2C9'] not in [
                           '*1/*1', '*1/*2', '*1/*3',
                           '*2/*2', '*2/*3', '*3/*3'])
                       # Not in EU-PACT
                       - 0.1092 * (patient['race'] == 'Asian')
                       # Not in EU-PACT
                       - 0.2760 * (patient['race'] == 'Black')
                       # # missing or mixed race - Not in EU-PACT
                       # - 1.0320 * (patient['race'] not in  [...])
                       # Enzyme inducer status(Fluvastatin is reductant
                       # not an inducer!) (comment by Ravvaz)
                       + 1.1816 * 0
                       - 0.5503 * (patient['amiodarone'] == 'Yes')) ** 2

        return weekly_dose / 7.0, 2

    @staticmethod
    def modified_pg(patient: Dict[str, Any]) -> Tuple[List[float], int]:
        '''
        Determine warfarin dose using the modified pharmacogenetic IWPC
        formula.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - height (in)
            - weight (lb)
            - VKORC1 ('G/G', 'G/A', 'A/A', etc.)
            - CYP2C9 ('*1/*1', '*1/*2', '*1/*3', '*2/*2', '*2/*3', '*3/*3',...)
            - amiodarone ('yes', 'no')

        Returns
        -------
        :
            A list of dose and time for the next test.

        Notes
        -----
        This method always returns 1 (day) for the next test.
        '''
        weekly_dose = (5.6044
                       # Based on Ravvaz (EU-PACT (page 18) uses year,
                       # Ravvaz (page 10 of annex) uses decades!
                       - 0.2614 * (patient['age'] / 10)
                       # in to cm
                       + 0.0087 * patient['height'] * 2.54
                       # lb to kg
                       + 0.0128 * patient['weight'] * 0.454
                       - 0.8677 * (patient['VKORC1'] == 'G/A')
                       - 1.6974 * (patient['VKORC1'] == 'A/A')
                       - 0.5211 * (patient['CYP2C9'] == '*1/*2')
                       - 0.9357 * (patient['CYP2C9'] == '*1/*3')
                       - 1.0616 * (patient['CYP2C9'] == '*2/*2')
                       - 1.9206 * (patient['CYP2C9'] == '*2/*3')
                       - 2.3312 * (patient['CYP2C9'] == '*3/*3')
                       - 0.5503 * (patient['amiodarone'] == 'Yes')) ** 2

        k = {'*1/*1': 0.0189,
             '*1/*2': 0.0158,
             '*1/*3': 0.0132,
             '*2/*2': 0.0130,
             '*2/*3': 0.0090,
             '*3/*3': 0.0075
             }
        LD3 = weekly_dose / ((1 - exp(-24*k[patient['CYP2C9']])) * (
            1 + exp(-24*k[patient['CYP2C9']]) + exp(-48*k[patient['CYP2C9']])))
        # The following dose calculation is based on EU-PACT report page 19
        # Ravvaz uses the same formula, but uses weekly dose. However,
        # EU-PACT explicitly mentions "predicted daily dose (D)"
        doses = [(1.5 * LD3 - 0.5 * weekly_dose) / 7,
                 LD3 / 7,
                 (0.5 * LD3 + 0.5 * weekly_dose) / 7]

        return doses, 1

    def reset(self) -> None:
        '''Reset the dosing protocol'''
        self._doses = []


class Lenzini(DosingProtocol):
    '''Lenzini warfarin dosing protocol based on Lenzini (2010).'''

    def prescribe(self,
                  patient: Dict[str, Any],
                  additional_info: Dict[str, Any]
                  ) -> Tuple[float, int, Dict[str, Any]]:
        if patient['day'] != 4:
            raise ValueError('Lenzini can only be called on day 4.')

        dose = exp(
            3.10894
            - 0.00767 * patient['age']
            - 0.51611 * log(patient['INRs'][-1])  # Natural log
            - 0.23032 * (1 * (patient['VKORC1'] == 'G/A')  # Heterozygous
                         + 2 * (patient['VKORC1'] == 'A/A'))  # Homozygous
            - 0.14745 * (1 * (patient['CYP2C9'] in ['*1/*2', '*2/*3'])  # Het.
                         + 2 * (patient['CYP2C9'] == '*2/*2'))  # Homozygous
            - 0.30770 * (1 * (patient['CYP2C9'] in ['*1/*3', '*2/*3'])  # Het.
                         + 2 * (patient['CYP2C9'] == '*3/*3'))  # Homozygous
            + 0.24597 * \
            sqrt(patient['height'] * 2.54 * \
                 patient['weight'] * 0.454 / 3600)  # BSA
            + 0.26729 * 2.5  # target INR
            - 0.10350 * (patient['amiodarone'] == 'Yes')
            + 0.01690 * patient['Doses'][-2]
            + 0.02018 * patient['Doses'][-3]
            # available if INR is measured on day 5
            + 0.01065 * patient['Doses'][-4]
        ) / 7

        return dose, 2, {}


class Intermountain(DosingProtocol):
    '''
    Intermountain warfarin dosing protocol based on `Anderson et al. (2007)
    supplements Appendix B
    <https://www.ahajournals.org/doi/10.1161/circulationaha.107.737312>`_
    '''

    def __init__(self, enforce_day_ge_8: bool = True) -> None:
        super().__init__()
        self._enforce_day_ge_8 = enforce_day_ge_8

    def prescribe(self,
                  patient: Dict[str, Any],
                  additional_info: Dict[str, Any]
                  ) -> Tuple[float, int, Dict[str, Any]]:
        dose_interval_list = additional_info.get('dose_interval_list', [])
        last_zone = additional_info.get('last_zone', '')
        previous_INR = patient['INRs'][-1]

        if not dose_interval_list:
            today = patient['day']

            if self._enforce_day_ge_8 and today < 8:
                raise ValueError('Intermountain is only valid for day>=8.')

            if self._enforce_day_ge_8 and today == 8:
                dose_list = patient['Doses']
                interval_list = patient['Intervals']
                dose_list = functools.reduce(
                    lambda x, y: x+y,
                    ([dose_list[-i]]*interval_list[-i]
                     for i in range(len(interval_list), 0, -1)))

                if len(dose_list) < 3:
                    raise ValueError(
                        'Intermountain requires doses for days 5 to 7 '
                        'for dosing on day 8.')

                previous_dose = sum(dose_list[-3:])/3
            else:
                previous_dose = patient['Doses'][-1]

            dose_interval_list, last_zone = self.intermountain_dosing_table(
                previous_INR, last_zone, previous_dose)

        else:
            if dose_interval_list[0].interval == -1:
                dose_interval_list, last_zone = \
                    self.intermountain_dosing_table(
                        previous_INR, last_zone, dose_interval_list[0].dose)

        additional_info['last_zone'] = last_zone
        additional_info['dose_interval_list'] = dose_interval_list[1:]

        return (dose_interval_list[0].dose,
                dose_interval_list[0].interval, additional_info)

    @staticmethod  # noqa: C901
    def intermountain_dosing_table(
            INR: float,
            last_zone: str,
            daily_dose: float) -> Tuple[List[DoseInterval], str]:
        '''
        Determine the dosing information, based on Intermountain dosing table.

        Arguments
        ---------
        current_INR:
            The latest value of INR.

        last_zone:
            The last zone the patient was in.
            * action point low
            * red zone low
            * yellow zone low
            * green zone
            * yellow zone high
            * red zone high
            * action point high

        daily_dose:
            The latest daily dose prescribed.

        Returns
        -------
        :
            * A list of `DoseInterval`s. It always includes the new daily dose
              and the new next test (in days). If an immediate dose is
              necessary, the first item will be the immediate dose and the next
              test day.
            * The new zone that patient's INR falls into.
        '''
        zone = Intermountain.zone(INR)

        weekly_dose = daily_dose * 7

        immediate_dose: float = -1.0
        immediate_interval: int = -1

        next_interval: int = {
            # -1 because immediate_interval = 1
            'action point low': (5-1, 14-1),
            'red zone low': (7-1, 14-1),  # -1 because immediate_interval = 1
            'yellow zone low': (14, 14),
            'green zone': (14, 28),
            'yellow zone high': (14, 14),
            'red zone high': (7-1, 14-1),  # -1 because immediate_interval = 1
            'action point high': (7, 14)
        }[zone][zone == last_zone]

        if last_zone == 'action point high':
            if zone in ['yellow zone low', 'green zone', 'yellow zone high']:
                weekly_dose *= 0.85
                next_interval = 7
            else:
                immediate_dose = 0.0
                immediate_interval = 2
                next_interval = -1

        elif zone == 'action point low':
            # (immediate extra dose) average 5-7 for day 8
            immediate_dose = daily_dose * 2
            immediate_interval = 1
            weekly_dose *= 1.10
        elif zone == 'red zone low':
            # (extra half dose) average 5-7 for day 8
            immediate_dose = daily_dose * 1.5
            immediate_interval = 1
            weekly_dose *= 1.05
        elif zone == 'yellow zone low':
            if zone == last_zone:
                weekly_dose *= 1.05
        # elif zone == 'green zone':
        #     pass
        elif zone == 'yellow zone high':
            if zone == last_zone:
                weekly_dose *= 0.95
        elif zone == 'red zone high':
            immediate_dose = daily_dose * 0.5 if INR < 4 else 0.0
            immediate_interval = 1
            weekly_dose *= 0.90
        elif zone == 'action point high':
            immediate_dose = 0.0
            immediate_interval = 2
            next_interval = -1

        dose_intervals = list(
            (DoseInterval(immediate_dose, immediate_interval),)
            if (immediate_dose >= 0.0 and immediate_interval > 0)
            else []
        ) + [DoseInterval(weekly_dose / 7, next_interval)]

        return dose_intervals, zone

    @staticmethod
    def zone(INR: float) -> str:
        '''
        Determine the zone based on patient's INR.

        Arguments
        ---------
        INR:
            the value of a patient's INR.

        Returns
        -------
        :
            Name of the dose, one of:
            * action point low
            * red zone low
            * yellow zone low
            * green zone
            * yellow zone high
            * red zone high
            * action point high
        '''
        if INR < 1.60:
            z = 'action point low'
        elif INR < 1.80:
            z = 'red zone low'
        elif INR < 2.00:
            z = 'yellow zone low'
        elif INR <= 3.00:
            z = 'green zone'
        elif INR < 3.40:
            z = 'yellow zone high'
        elif INR < 5.00:
            z = 'red zone high'
        else:
            z = 'action point high'

        return z


class CompositeDosingProtocol:
    '''
    A dosing protocol class that can contain three dosing protocols for
    `initial`, `adjustment` and `maintenance` phases of dosing.
    '''

    def __init__(self,
                 initial_protocol: DosingProtocol,
                 adjustment_protocol: DosingProtocol,
                 maintenance_protocol: DosingProtocol) -> None:
        '''
        Arguments
        ---------
        initial_protocol:
            A dosing protocol for the initial phase of dosing.

        adjustment_protocol
            A dosing protocol for the adjustment phase of dosing.

        maintenance_protocol
            A dosing protocol for the maintenance phase of dosing.
        '''
        self._initial_protocol = initial_protocol
        self._adjustment_protocol = adjustment_protocol
        self._maintenance_protocol = maintenance_protocol
        self._additional_info = {}

    def prescribe(self, patient: Dict[str, Any]) -> Tuple[float, int]:
        '''
        Prescribe a dose and next test (in days) for the given `patient`.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics necessary to make dosing
            decisions.

        Returns
        -------
        :
            The prescribed dose and the time of the next test (in days).
        '''
        raise NotImplementedError

    def reset(self) -> None:
        '''Reset the dosing protocol.'''
        self._initial_protocol.reset()
        self._adjustment_protocol.reset()
        self._maintenance_protocol.reset()
        self._additional_info = {}


class AAA(CompositeDosingProtocol):
    '''
    A composite dosing protocol with `Aurora` in all phases.
    '''

    def __init__(self) -> None:
        aurora_instance = Aurora()
        super().__init__(aurora_instance, aurora_instance, aurora_instance)

    def prescribe(self, patient: Dict[str, Any]) -> Tuple[float, int]:
        dose, interval, self._additional_info = \
            self._initial_protocol.prescribe(patient, self._additional_info)

        return dose, interval


class CAA(CompositeDosingProtocol):
    '''
    A composite dosing protocol with clinical `IWPC` in initial phase, and
    `Aurora` in adjustment and maintenance phases.
    '''

    def __init__(self) -> None:
        iwpc_instance = IWPC('clinical')
        aurora_instance = Aurora()
        super().__init__(iwpc_instance, aurora_instance, aurora_instance)

    def prescribe(self, patient: Dict[str, Any]) -> Tuple[float, int]:
        fn = (self._initial_protocol if patient['day'] <= 2
              else self._adjustment_protocol)
        dose, interval, self._additional_info = fn.prescribe(
            patient, self._additional_info)

        return dose, interval


class PGAA(CompositeDosingProtocol):
    '''
    A composite dosing protocol with pharmacogenetic `IWPC` in initial phase,
    and `Aurora` in adjustment and maintenance phases.
    '''

    def __init__(self) -> None:
        iwpc_instance = IWPC('pharmacogenetic')
        aurora_instance = Aurora()
        super().__init__(iwpc_instance, aurora_instance, aurora_instance)

    def prescribe(self, patient: Dict[str, Any]) -> Tuple[float, int]:
        fn = (self._initial_protocol if patient['day'] <= 2
              else self._adjustment_protocol)
        dose, interval, self._additional_info = fn.prescribe(
            patient, self._additional_info)

        return dose, interval


class PGPGA(CompositeDosingProtocol):
    '''
    A composite dosing protocol with modified `IWPC` in initial phase,
    `Lenzini` in adjustment phase, and `Aurora` in maintenance phase.
    '''

    def __init__(self) -> None:
        iwpc_instance = IWPC('modified')
        lenzini_instance = Lenzini()
        aurora_instance = Aurora()
        super().__init__(iwpc_instance, lenzini_instance, aurora_instance)

    def prescribe(self, patient: Dict[str, Any]) -> Tuple[float, int]:
        if patient['day'] <= 3:
            fn = self._initial_protocol
        elif patient['day'] <= 5:
            fn = self._adjustment_protocol
        else:
            fn = self._maintenance_protocol

        dose, interval, self._additional_info = fn.prescribe(
            patient, self._additional_info)

        return dose, interval


class PGPGI(CompositeDosingProtocol):
    '''
    A composite dosing protocol with modified `IWPC` in initial phase,
    `Lenzini` in adjustment phase, and `Intermountain` in maintenance phase.
    '''

    def __init__(self) -> None:
        iwpc_instance = IWPC('modified')
        lenzini_instance = Lenzini()
        intermountain_instance = Intermountain(enforce_day_ge_8=False)
        super().__init__(
            iwpc_instance, lenzini_instance, intermountain_instance)

    def prescribe(self, patient: Dict[str, Any]) -> Tuple[float, int]:
        if patient['day'] <= 3:
            fn = self._initial_protocol
        elif patient['day'] <= 5:
            fn = self._adjustment_protocol
        else:
            fn = self._maintenance_protocol

        dose, interval, self._additional_info = fn.prescribe(
            patient, self._additional_info)

        return dose, interval


class WarfarinAgent(agents.NoLearnAgent):
    '''
    An `agent` that prescribes dose for a warfarin `subject`,
    based on the dosing protocols defined in Ravvaz et al (2017).
    '''

    def __init__(self,
                 study_arm: Literal['aaa', 'caa', 'pgaa',
                                    'pgpgi', 'pgpga'] = 'aaa',
                 **kwargs: Any):
        '''
        Arguments
        ---------
        study_arm:
            One of available study arms: aaa, caa, pgaa, pgpgi, pgpga
        '''
        super().__init__(**kwargs)

        if study_arm.lower() in ['aaa', 'ravvaz aaa', 'ravvaz_aaa']:
            self._protocol = AAA()
        elif study_arm.lower() in ['caa', 'ravvaz caa', 'ravvaz_caa']:
            self._protocol = CAA()
        elif study_arm.lower() in ['pgaa', 'ravvaz pgaa', 'ravvaz_pgaa']:
            self._protocol = PGAA()
        elif study_arm.lower() in ['pgpgi', 'ravvaz pgpgi', 'ravvaz_pgpgi']:
            self._protocol = PGPGI()
        elif study_arm.lower() in ['pgpga', 'ravvaz pgpga', 'ravvaz_pgpga']:
            self._protocol = PGPGA()

    def act(self,
            state: ReilData,
            actions: Optional[Tuple[ReilData, ...]] = None,
            epoch: Optional[int] = 0) -> ReilData:
        '''
        Generate the dosing `action` based on the `state` and current dosing
        protocol.

        Arguments
        ---------
        state:
            The state for which the action should be returned.

        actions:
            The set of possible actions to choose from.

        epoch:
            The epoch in which the agent is acting.

        Returns
        -------
        :
            The action
        '''
        patient = state.value
        patient['day'] += 1

        dose, interval = self._protocol.prescribe(patient)

        # Cuts out the dose if it is >15.0
        return ReilData([
            {'name': 'dose', 'value': min(dose, 15.0),
             'lower': 0.0, 'upper': 15.0},
            {'name': 'interval', 'value': min(interval, 28),
             'lower': 1, 'upper': 28}
        ])

    def reset(self):
        '''Resets the agent at the end of a learning epoch.'''
        self._protocol.reset()

    def __repr__(self) -> str:
        try:
            return super().__repr__() + f' arm: {self._protocol}'
        except NameError:
            return super().__repr__()

# Implementation that matches RAVVAZ dataset.
# -------------------------------------------
#
# def aurora(self, patient: Dict[str, Any]) -> float:
#     if self._red_flag:
#         if self._retest_day > patient['day']:
#             return_value = 0.0
#         elif patient['INRs'][-1] > 3.0:
#             self._retest_day = patient['day'] + 2
#             return_value = 0.0
#         else:
#             self._red_flag = False
#             self._retest_day = patient['day'] + 7
#             return_value = self._dose
#         return return_value

#     next_test = 2
#     if patient['day'] <= 2:
#         self._dose = 10.0 if patient['age'] < 65.0 else 5.0
#     elif patient['day'] <= 4:
#         day_2_INR = (patient['INRs'][-1] if patient['day'] == 3
#                      else patient['INRs'][-2])
#         if day_2_INR >= 2.0:
#             self._dose = 5.0
#             if day_2_INR <= 3.0:
#                 self._early_therapeutic = True

#             self._number_of_stable_days, next_test = \
#                 self._aurora_retesting_table(patient['INRs'][-1],
#                                              self._number_of_stable_days,
#                                              self._early_therapeutic)
#         else:
#             self._dose, next_test, _, _ = self._aurora_dosing_table(
#                 day_2_INR, self._dose)
#     else:
#         self._number_of_stable_days, next_test = \
#             self._aurora_retesting_table(
#               patient['INRs'][-1],
#               self._number_of_stable_days,
#               self._early_therapeutic)

#     if next_test == -1:
# #         self._early_therapeutic = False
#         self._number_of_stable_days = 0
#         self._dose, next_test, self._skip_dose, self._red_flag = \
#           self._aurora_dosing_table(patient['INRs'][-1], self._dose)

#     self._retest_day = patient['day'] + next_test

#     return self._dose if self._skip_dose == 0 else 0.0

# def _aurora_dosing_table(self,
#     current_INR: float, dose: float) -> Tuple[float, int, int, bool]:
#     skip_dose = 0
#     red_flag = False
#     if current_INR < 1.50:
#         dose = dose * 1.15
#         next_test = 7
#     elif current_INR < 1.80:
#         dose = dose * 1.10
#         next_test = 7
#     elif current_INR < 2.00:
#         dose = dose * 1.075
#         next_test = 7
#     elif current_INR <= 3.00:
#         next_test = 28
#     elif current_INR < 3.40:
#         dose = dose * 0.925
#         next_test = 7
#     elif current_INR < 4.00:
#         dose = dose * 0.9
#         next_test = 7
#     elif current_INR <= 5.00:
#         skip_dose = 1  # 2
#         dose = dose * 0.875
#         next_test = 7
#     else:
#         red_flag = True
#         next_test = 2
#         dose = dose * 0.85

#     return dose, next_test, skip_dose, red_flag

# def _aurora_retesting_table(self,
#   current_INR: float, number_of_stable_days: int,
#   early_therapeutic: bool = False) -> Tuple[int, int]:
#     # next_test = {0: 1, 1: 1, 2: 5, 7: 7, 14: 14, 28: 28}
#     # NOTE: When patient gets into the range early on
#     #      (before the maintenance period), we have 1: 2,
#     # but when patient is in maintenance,\
#     # we have 1: 6. I track the former in the main aurora method.

#     if early_therapeutic:
#         next_test = {0: 1, 1: 2, 3: 4, 7: 6, 13: 6, 19: 13, 32: 27}
#         max_gap = 32
#     else:
#         next_test = {0: 1, 1: 6, 7: 6, 13: 13, 26: 27}
#         max_gap = 26
#     if 2.0 <= current_INR <= 3.0:
#         number_of_stable_days = min(
#             number_of_stable_days + next_test[number_of_stable_days],
#             max_gap)
#     else:
#         return -1, -1

#     return number_of_stable_days, next_test[number_of_stable_days]



# -*- coding: utf-8 -*-
'''
VanillaExperienceReplay class
=============================

A `Buffer` with random pick that picks only if it is full.
'''

from typing import Dict, List, Optional, Tuple

from reil.datatypes.buffers import CircularBuffer, T


class VanillaExperienceReplay(CircularBuffer[T]):
    '''
    A `Buffer` with random pick that picks only if it is full.

    Extends `CircularBuffer` class.
    '''
    _batch_size = None
    _clear_buffer = False

    def __init__(self,
                 buffer_size: Optional[int] = None,
                 batch_size: Optional[int] = None,
                 buffer_names: Optional[List[str]] = None,
                 clear_buffer: bool = False) -> None:
        '''
        Initialize the buffer.

        Arguments
        -----------
        buffer_size:
            The size of the buffer.

        batch_size:
            The number of items to return at each `pick`.

        buffer_names:
            A list containing the names of buffer queues.

        clear_buffer:
            Whether to clear the buffer when `reset` is called.
        '''
        self.setup(buffer_size=buffer_size,
                   batch_size=batch_size,
                   buffer_names=buffer_names,
                   clear_buffer=clear_buffer)

        super().setup(pick_mode='random')

    def setup(self,
              buffer_size: Optional[int] = None,
              batch_size: Optional[int] = None,
              buffer_names: Optional[List[str]] = None,
              clear_buffer: bool = False) -> None:
        '''
        Set up the buffer.

        Arguments
        ---------
        buffer_size:
            The size of the buffer.

        batch_size:
            The number of items to return at each `pick`.

        buffer_names:
            A list containing the names of buffer elements.

        clear_buffer:
            Whether to clear the buffer when `reset` is called.

        Notes
        -----
        `setup` should be used only for attributes of the buffer that are
        not defined. Attempt to use `setup` to modify size, names or mode will
        result in an exception.
        '''
        if buffer_size is not None and buffer_size < 1:
            raise ValueError('buffer_size should be at least 1.')

        super().setup(buffer_size=buffer_size, buffer_names=buffer_names)

        if self._buffer_size is not None:
            if batch_size is not None and self._buffer_size < batch_size:
                raise ValueError('buffer_size should be >= batch_size.')

        if batch_size is not None:
            if self._batch_size is not None:
                raise ValueError(
                    'Cannot modify batch_size. The value is already set.')
            else:
                self._batch_size = batch_size

        self._clear_buffer = clear_buffer

        self.reset()

    def pick(self) -> Dict[str, Tuple[T, ...]]:
        '''
        Return `batch_size` number of items from the buffer randomly.
        If the buffer is not full, return empty tuples.
        '''
        if self._buffer_full:
            return super().pick(self._batch_size, 'random')
        else:
            return dict((name, ()) for name in self._buffer)

    def reset(self) -> None:
        '''
        Reset the buffer if `clear_buffer` is set to `True`.
        '''
        if self._clear_buffer:
            super().reset()
            self._buffer_full = False



# -*- coding: utf-8 -*-
'''
EnvironmentStaticMap class
==========================

This class provides a learning environment for any reinforcement learning
`agent` on any `subject`. The interactions between `agents` and `subjects`
are determined by a fixed `interaction_sequence`.
'''
from typing import Any, Dict, Optional, Tuple, Union, cast

from reil import agents as rlagents
from reil import environments
from reil import subjects as rlsubjects
from reil.datatypes import InteractionProtocol
from reil.utils import instance_generator as rlgenerator

AgentSubjectTuple = Tuple[str, str]
Entity = Union[rlagents.Agent, rlsubjects.Subject]
EntityGenerator = Union[rlgenerator.InstanceGenerator[rlagents.Agent],
                        rlgenerator.InstanceGenerator[rlsubjects.Subject]]


class EnvironmentStaticMap(environments.Environment):
    '''
    Provide an interaction and learning environment for `agents` and
    `subjects`, based on a static interaction sequence.
    '''

    def __init__(self,
                 entity_dict: Optional[
                     Dict[str, Union[Entity, EntityGenerator, str]]] = None,
                 interaction_sequence: Optional[
                     Tuple[InteractionProtocol, ...]] = None,
                 **kwargs: Any):
        '''
        Arguments
        ---------
        entity_dict:
            a dictionary that contains `agents`, `subjects`, and
            `generators`.

        interaction_sequence:
            a tuple of `InteractionProtocols` that specify
            how entities interact in the simulation.
        '''
        super().__init__(entity_dict=entity_dict, **kwargs)

        self._interaction_sequence: Tuple[InteractionProtocol, ...] = ()

        if interaction_sequence is not None:
            self.interaction_sequence = interaction_sequence

    def remove(self, entity_names: Tuple[str, ...]) -> None:
        '''
        Extends `Environment.remove`.

        Remove `agents`, `subjects`, or `instance_generators` from
        the environment.

        Arguments
        ---------
        entity_names:
            A list of `agent`/ `subject` names to be deleted.

        Raises
        ------
        RuntimeError
            The entity listed for deletion is used in the
            `interaction_sequence`.

        Notes
        -----
        This method removes the item from both `agents` and `subjects`
        lists. Hence, it is not recommended to use the same name for both
        an `agent` and a `subject`.
        '''
        names_in_use = [p.agent.name
                        for p in self._interaction_sequence] + \
                       [p.subject.name
                        for p in self._interaction_sequence]
        temp = set(entity_names).difference(names_in_use)
        if temp:
            raise RuntimeError(f'Some entities are in use: {temp}')

        super().remove(entity_names)

    @property
    def interaction_sequence(self) -> Tuple[InteractionProtocol, ...]:
        return self._interaction_sequence

    @interaction_sequence.setter
    def interaction_sequence(self,
                             seq: Tuple[InteractionProtocol, ...]) -> None:
        self._agent_observers = {}
        for protocol in seq:
            self.assert_protocol(protocol)
            self.register(protocol, get_agent_observer=True)

        self._interaction_sequence = seq

    def simulate_one_pass(self) -> None:
        '''
        Go through the interaction sequence for one pass and
        simulate interactions accordingly.
        '''
        for protocol in self._interaction_sequence:
            subject_name = protocol.subject.name

            if self._subjects[subject_name].is_terminated(None):
                continue

            agent_name = protocol.agent.name
            a_s_name = (agent_name, subject_name)
            unit = protocol.unit
            state_name = protocol.state_name
            reward_function_name = protocol.reward_function_name
            agent_id, _ = cast(Tuple[int, int],
                               self._assignment_list[a_s_name])

            if unit == 'interaction':
                if protocol.n == 1:
                    self.interact_once(
                        agent_id=agent_id,
                        agent_observer=self._agent_observers[a_s_name],
                        subject_instance=self._subjects[subject_name],
                        state_name=state_name,
                        reward_function_name=reward_function_name,
                        epoch=self._epochs[subject_name])
                else:
                    self.interact_n_times(
                        agent_id=agent_id,
                        agent_observer=self._agent_observers[a_s_name],
                        subject_instance=self._subjects[subject_name],
                        state_name=state_name,
                        reward_function_name=reward_function_name,
                        epoch=self._epochs[subject_name],
                        times=protocol.n)

                if self._subjects[subject_name].is_terminated(None):
                    self.check_subject(subject_name)

            elif unit in ['instance', 'epoch']:
                # For epoch, simulate the current instance, then in the next if
                # statement, simulate the rest of the generated instances.
                self.interact_while(
                    agent_id=agent_id,
                    agent_observer=self._agent_observers[a_s_name],
                    subject_instance=self._subjects[subject_name],
                    state_name=state_name,
                    reward_function_name=reward_function_name,
                    epoch=self._epochs[subject_name])

                if (unit == 'epoch'
                        and subject_name in self._instance_generators):
                    while self.check_subject(subject_name):
                        self.interact_while(
                            agent_id=agent_id,
                            agent_observer=self._agent_observers[a_s_name],
                            subject_instance=self._subjects[subject_name],
                            state_name=state_name,
                            reward_function_name=reward_function_name,
                            epoch=self._epochs[subject_name])

                else:
                    self.check_subject(subject_name)

            else:
                raise ValueError(f'Unknown protocol unit: {unit}.')

    def simulate_to_termination(self) -> None:
        '''
        Go through the interaction sequence and simulate interactions
        accordingly, until all `subjects` are terminated.

        Notes
        -----
        To avoid possible infinite loops caused by normal `subjects`,
        this method is only available if all `subjects` are generated
        by `instance generators`.

        Raises
        ------
        TypeError:
            Attempt to call this method will normal subjects in the interaction
            sequence.
        '''
        subjects_in_use = set(s.subject.name
                              for s in self.interaction_sequence)
        no_generators = subjects_in_use.difference(self._instance_generators)
        if no_generators:
            raise TypeError(
                'Found subject(s) in the interaction_sequence that '
                f'are not instance generators: {no_generators}')

        infinites = [s
                     for s in subjects_in_use
                     if not self._instance_generators[s].is_finite]
        if infinites:
            raise TypeError('Found infinite instance generator(s) in the '
                            f'interaction_sequence: {infinites}')

        while not all(self._instance_generators[s].is_terminated()
                      for s in self._subjects):
            self.simulate_one_pass()

        self.report_statistics(True)

    def check_subject(self, subject_name: str) -> bool:
        '''
        Go over all `subjects`. If terminated, close related `agent_observers`,
        reset the `subject`, and create new `agent_observers`.
        '''
        # print(self._subjects[subject_name])
        affected_protocols = list(p for p in self._interaction_sequence
                                  if p.subject.name == subject_name)

        success: bool = True
        if affected_protocols:
            for p in affected_protocols:
                self.close_agent_observer(p)

            success = self.reset_subject(subject_name)

            for p in affected_protocols:
                self.register(p, get_agent_observer=True)

        return success

    def reset_subject(self, subject_name: str) -> bool:
        '''
        Extends `Environment.reset_subject()`.
        '''
        entities = set(
            (p.subject.statistic_name,
             self._assignment_list[(p.agent.name, p.subject.name)][0])
            for p in self.interaction_sequence
            if p.subject.name == subject_name and
            p.subject.statistic_name is not None)

        for e in entities:
            if subject_name in self._instance_generators:
                self._instance_generators[subject_name].statistic.append(*e)
            else:
                self._subjects[subject_name].statistic.append(*e)

        return super().reset_subject(subject_name)

    def report_statistics(self, reset_history: bool = False):
        entities = set(
            (p.subject.name, p.subject.aggregators, p.subject.groupby,
             self._assignment_list[(p.agent.name, p.subject.name)][0])
            for p in self.interaction_sequence
            if p.subject.statistic_name is not None and
            p.subject.aggregators is not None)

        for e in entities:
            print(e)
            if e[0] in self._instance_generators:
                print(self._instance_generators[e[0]].statistic.aggregate(
                    e[1], e[2], e[3], reset_history=reset_history))
            else:
                print(self._subjects[e[0]].statistic.aggregate(
                    e[1], e[2], e[3], reset_history=reset_history))

        entities = set(
            (p.agent.name, p.agent.aggregators, p.agent.groupby,
             self._assignment_list[(p.agent.name, p.subject.name)][1])
            for p in self.interaction_sequence
            if p.agent.statistic_name is not None and
            p.agent.aggregators is not None)

        for e in entities:
            print(e)
            if e[0] in self._instance_generators:
                print(self._instance_generators[e[0]].statistic.aggregate(
                    e[1], e[2], e[3], reset_history=reset_history))
            else:
                print(self._agents[e[0]].statistic.aggregate(
                    e[1], e[2], e[3], reset_history=reset_history))



# -*- coding: utf-8 -*-
'''
Dense class
===========

The Dense learner.
'''
import pathlib
from typing import Any, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
from reil import learners
from reil.datatypes import ReilData
from tensorflow import keras  # type: ignore


class Dense(learners.Learner[float]):
    '''
    The Dense learner.

    This class uses `tf.keras` to build a sequential dense network with one
    output.
    '''

    def __init__(self,
                 learning_rate: learners.LearningRateScheduler,
                 validation_split: float = 0.3,
                 hidden_layer_sizes: Tuple[int, ] = (1,),
                 input_length: Optional[int] = None,
                 tensorboard_path: Optional[Union[str, pathlib.Path]] = None,
                 **kwargs: Any) -> None:
        '''
        Arguments
        ---------
        learning_rate:
            A `LearningRateScheduler` object that determines the learning rate
            based on epoch. If any scheduler other than constant is provided,
            the model uses the `new_rate` method of the scheduler to determine
            the learning rate at each epoch.

        validation_split:
            How much of the training set should be used for validation?

        hidden_layer_sizes:
            A list of number of neurons for each layer.

        input_length:
            Size of the input data. If not supplied, the network will be
            generated based on the size of the first data point in `predict` or
            `learn` methods.

        tensorboard_path:
            A path to save tensorboard outputs. If not provided,
            tensorboard will be disabled.

        Raises
        ------
        ValueError
            Validation split not in the range of (0.0, 1.0).
        '''

        super().__init__(learning_rate=learning_rate, **kwargs)

        self._epoch = 0

        self._hidden_layer_sizes = hidden_layer_sizes
        self._input_length = input_length

        if not 0.0 < validation_split < 1.0:
            raise ValueError('validation split should be in (0.0, 1.0).')
        self._validation_split = validation_split

        self._tensorboard_path = tensorboard_path

        if self._input_length is not None:
            self._generate_network()
        else:
            self._graph = None

    def _generate_network(self) -> None:
        '''
        Generate a multilayer neural net using `keras.Dense`.
        '''

        self._graph = tf.Graph()  # type: ignore
        with self._graph.as_default():
            self._session = tf.Session()  # type: ignore

            self._model = keras.models.Sequential()
            self._model.add(
                keras.layers.Dense(self._hidden_layer_sizes[0],
                                   activation='relu',
                                   name='layer_01',
                                   input_shape=(self._input_length,)))
            for i, v in enumerate(self._hidden_layer_sizes[1:]):
                self._model.add(keras.layers.Dense(
                    v, activation='relu', name=f'layer_{i+2:0>2}'))

            self._model.add(keras.layers.Dense(1, name='output'))

            self._model.compile(optimizer=keras.optimizers.Adam(
                learning_rate=self._learning_rate.initial_lr), loss='mae')

            self._callbacks = []
            if self._tensorboard_path is not None:
                self._tensorboard_path = pathlib.Path(
                    'logs', self._tensorboard_path)
                self._tensorboard = keras.callbacks.TensorBoard(
                    log_dir=self._tensorboard_path)
                # , histogram_freq=1)  #, write_images=True)
                self._callbacks.append(self._tensorboard)

            if not isinstance(self._learning_rate,
                              learners.ConstantLearningRate):
                self._learning_rate_scheduler = \
                    keras.callbacks.LearningRateScheduler(
                        self._learning_rate.new_rate, verbose=0)
                self._callbacks.append(self._learning_rate_scheduler)

    def predict(self, X: Tuple[ReilData, ...]) -> Tuple[float, ...]:
        '''
        predict `y` for a given input list `X`.

        Arguments
        ---------
        X:
            A list of `ReilData` as inputs to the prediction model.

        Returns
        -------
        :
            The predicted `y`.
        '''
        _X = [x.normalized.flatten() for x in X]
        if self._graph is None:
            self._input_length = len(_X[0])
            self._generate_network()

        with self._session.as_default():
            with self._graph.as_default():
                result = self._model.predict(np.array(_X))

        return result

    def learn(self, X: Tuple[ReilData, ...], Y: Tuple[float, ...]) -> None:
        '''
        Learn using the training set `X` and `Y`.

        Arguments
        ---------
        X:
            A list of `ReilData` as inputs to the learning model.

        Y:
            A list of float labels for the learning model.
        '''
        _X = [x.normalized.flatten() for x in X]
        if self._graph is None:
            self._input_length = len(_X[0])
            self._generate_network()

        with self._session.as_default():
            with self._graph.as_default():
                self._model.fit(
                    np.array(_X), np.array(Y),
                    initial_epoch=self._epoch, epochs=self._epoch+1,
                    callbacks=self._callbacks,
                    validation_split=self._validation_split,
                    verbose=0)

    def reset(self) -> None:
        '''
        reset the learner.
        '''
        self._epoch += 1

    def save(self,
             filename: str,
             path: pathlib.Path) -> Tuple[pathlib.Path, str]:
        '''
        Extends `ReilBase.save` to handle `TF` objects.

        Arguments
        ---------
        filename:
            The name of the file to be saved.

        path:
            The path of the file to be saved.

        Raises
        ------
        ValueError:
            The filename is not specified.
        '''
        path.mkdir(parents=True, exist_ok=True)
        with self._session.as_default():
            with self._graph.as_default():
                self._model.save(path / filename)

        return path, filename

    def load(self,
             filename: str,
             path: Optional[Union[str, pathlib.Path]] = None) -> None:
        '''
        Extends `ReilBase.load` to handle `TF` objects.

        Arguments
        ---------
        filename:
            The name of the file to be loaded.

        path:
            Path of the location of the file.

        Raises
        ------
        ValueError:
            The filename is not specified.
        '''
        _path = path if path is not None else '.'
        self._graph = tf.Graph()  # type: ignore
        with self._graph.as_default():
            self._session = keras.backend.get_session()
            self._model = keras.models.load_model(
                pathlib.Path(_path, f'{filename}.tf', filename))
            self._tensorboard = keras.callbacks.TensorBoard(
                log_dir=self._tensorboard_path)
            # , histogram_freq=1)  # , write_images=True)
            self._learning_rate_scheduler = \
                keras.callbacks.LearningRateScheduler(
                    self._learning_rate_scheduler)



# -*- coding: utf-8 -*-  pylint: disable=undefined-variable
'''
warfarin class
==============

This `warfarin` class implements a two compartment PK/PD model for warfarin.
'''

import pathlib
from typing import Any, Dict, List, Optional, Tuple

from reil.datatypes import ReilData
from reil.subjects import Subject, healthcare
from reil.utils import action_generator, reil_functions


class Warfarin(Subject):
    '''
    A warfarin subject based on Hamberg's two compartment PK/PD model.
    '''

    def __init__(self,
                 patient: healthcare.Patient,
                 action_generator: action_generator.ActionGenerator,
                 max_day: int = 90,
                 **kwargs: Any):
        '''
        Arguments
        ---------
        patient:
            A patient object that generates new patients and models
            interaction between dose and INR.

        action_generator:
            An `ActionGenerator` object with 'dose' and
            'interval' components.

        max_day:
            Maximum duration of each trial.

        Raises
        ------
        ValueError
            action_generator should have a "dose" component.

        ValueError
            action_generator should have an "interval" component.
        '''

        super().__init__(max_agent_count=1, **kwargs)

        if 'dose' not in action_generator.components:
            raise ValueError(
                'action_generator should have a "dose" component.')
        if 'interval' not in action_generator.components:
            raise ValueError(
                'action_generator should have an "interval" component.')

        self._patient = patient
        self._action_generator = action_generator
        self._max_day = max_day
        patient_basic = (('age', {}), ('CYP2C9', {}),
                         ('VKORC1', {}), ('sensitivity', {}))
        patient_extra = (('weight', {}), ('height', {}),
                         ('gender', {}), ('tobaco', {}),
                         ('amiodarone', {}), ('fluvastatin', {}))

        reward_sq_dist = reil_functions.NormalizedSquareDistance(
            name='sq_dist', arguments=('daily_INR',),  # type: ignore
            length=-1, multiplier=-1.0, retrospective=True, interpolate=False,
            center=2.5, band_width=1.0, exclude_first=True)

        reward_sq_dist_interpolation = reil_functions.NormalizedSquareDistance(
            name='sq_dist_interpolation',
            arguments=('INR', 'interval'),  # type: ignore
            length=2, multiplier=-1.0, retrospective=True, interpolate=True,
            center=2.5, band_width=1.0, exclude_first=True)

        reward_PTTR = reil_functions.PercentInRange(
            name='PTTR', arguments=('daily_INR',),  # type: ignore
            length=-1, multiplier=-1.0, retrospective=True, interpolate=False,
            acceptable_range=(2, 3), exclude_first=True)

        reward_PTTR_interpolation = reil_functions.PercentInRange(
            name='PTTR', arguments=('INR', 'interval'),  # type: ignore
            length=2, multiplier=-1.0, retrospective=True, interpolate=True,
            acceptable_range=(2, 3), exclude_first=True)

        statistic_PTTR = reil_functions.PercentInRange(
            name='PTTR', arguments=('daily_INR',),  # type: ignore
            length=-1, multiplier=1.0, retrospective=True, interpolate=False,
            acceptable_range=(2, 3), exclude_first=True)

        self.state.add_definition('patient_basic',
                                  *patient_basic)

        self.state.add_definition('patient',
                                  *patient_basic,
                                  *patient_extra)

        self.state.add_definition('patient_w_dosing',
                                  *patient_basic,
                                  *patient_extra,
                                  ('day', {}),
                                  ('dose', {'length': -1}),
                                  ('INR', {'length': -1}),
                                  ('interval', {'length': -1}))

        for i in (1, 5, 10):
            self.state.add_definition(f'patient_w_dosing_{i:02}',
                                      *patient_basic,
                                      *patient_extra,
                                      ('day', {}),
                                      ('dose', {'length': i}),
                                      ('INR', {'length': i}),
                                      ('interval', {'length': i}))

        self.state.add_definition('patient_w_full_dosing',
                                  *patient_basic,
                                  *patient_extra,
                                  ('day', {}),
                                  ('daily_dose', {'length': -1}),
                                  ('daily_INR', {'length': -1}),
                                  ('interval', {'length': -1}))

        self.state.add_definition('daily_INR',
                                  ('daily_INR', {'length': -1}))

        self.state.add_definition('Measured_INR_2',
                                  ('INR', {'length': 2}),
                                  ('interval', {'length': 1}))

        self.state.add_definition('INR_within_2',
                                  ('INR_within', {'length': 1}))

        self.reward.add_definition(
            'no_reward', lambda _: 0.0, 'Measured_INR_2')

        self.reward.add_definition(
            'sq_dist_exact', reward_sq_dist, 'INR_within_2')

        self.reward.add_definition(
            'sq_dist_interpolation', reward_sq_dist_interpolation,
            'Measured_INR_2')

        self.reward.add_definition(
            'PTTR_exact', reward_PTTR, 'INR_within_2')

        self.reward.add_definition(
            'PTTR_interpolation', reward_PTTR_interpolation, 'Measured_INR_2')

        self.statistic.add_definition(
            'PTTR_exact_basic', statistic_PTTR, 'daily_INR', 'patient_basic')

        self.statistic.add_definition(
            'PTTR_exact', statistic_PTTR, 'daily_INR', 'patient')

        self.reset()

    @staticmethod
    def generate_dose_actions(min_dose: float = 0.0,
                              max_dose: float = 15.0,
                              dose_increment: float = 0.5) -> List[float]:

        return list(min_dose + x * dose_increment
                    for x in range(
                        int((max_dose - min_dose)/dose_increment) + 1))

    @staticmethod
    def generate_interval_actions(min_interval: int = 1,
                                  max_interval: int = 28,
                                  interval_increment: int = 1) -> List[int]:

        return list(range(min_interval, max_interval, interval_increment))

    def is_terminated(self, _id: Optional[int] = None) -> bool:
        return self._day >= self._max_day

    def possible_actions(
            self, _id: Optional[int] = None) -> Tuple[ReilData, ...]:
        return self._action_generator.possible_actions(
            self.state('default', _id))

    def take_effect(self,
                    action: ReilData,
                    _id: int = 0) -> None:
        Subject.take_effect(self, action, _id)
        current_dose = float(action.value['dose'])
        current_interval = min(int(action.value['interval']),
                               self._max_day - self._day)

        INRs_temp = self._patient.model(
            dose=dict(((i + self._day, current_dose)
                       for i in range(current_interval))),
            measurement_days=list(
                range(self._day + 1, self._day + current_interval + 1)))['INR']

        self._decision_points_dose_history[self._decision_points_index] = \
            current_dose
        self._decision_points_interval_history[self._decision_points_index] = \
            current_interval
        self._decision_points_index += 1

        day_temp = self._day
        self._day += current_interval

        self._full_dose_history[day_temp:self._day] = \
            [current_dose] * current_interval
        self._full_INR_history[day_temp + 1:self._day + 1] = INRs_temp

        self._decision_points_INR_history[self._decision_points_index] = \
            self._full_INR_history[self._day]

    def reset(self) -> None:
        Subject.reset(self)
        self._patient.generate()

        self._day: int = 0
        self._full_INR_history = [0.0] * self._max_day
        self._full_dose_history = [0.0] * self._max_day
        self._decision_points_INR_history = [0.0] * (self._max_day + 1)
        self._decision_points_dose_history = [0.0] * self._max_day
        self._decision_points_interval_history: List[int] = [1] * self._max_day
        self._decision_points_index: int = 0

        self._full_INR_history[0] = self._patient.model(
            measurement_days=[0])['INR'][-1]
        self._decision_points_INR_history[0] = self._full_INR_history[0]

    def _default_state_definition(
            self, _id: Optional[int] = None) -> ReilData:
        patient_features = self._patient.feature_set
        return ReilData([
            {'name': 'age',
             'value': patient_features['age'].value,
             'lower': patient_features['age'].lower,
             'upper': patient_features['age'].upper},
            {'name': 'CYP2C9',
             'value': patient_features['CYP2C9'].value,
             'categories': patient_features['CYP2C9'].categories},
            {'name': 'VKORC1',
             'value': patient_features['VKORC1'].value,
             'categories': patient_features['VKORC1'].categories}],
            lazy_evaluation=True)

    def _numerical_sub_comp(self, name):
        temp = self._patient.feature_set[name]
        return {'name': name,
                'value': temp.value,
                'lower': temp.lower,
                'upper': temp.upper}

    def _categorical_sub_comp(self, name):
        temp = self._patient.feature_set[name]
        return {'name': name,
                'value': temp.value,
                'categories': temp.categories}

    def _sub_comp_age(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return self._numerical_sub_comp('age')

    def _sub_comp_weight(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return self._numerical_sub_comp('weight')

    def _sub_comp_height(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return self._numerical_sub_comp('height')

    def _sub_comp_gender(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return self._categorical_sub_comp('gender')

    def _sub_comp_race(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return self._categorical_sub_comp('race')

    def _sub_comp_tobaco(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return self._categorical_sub_comp('tobaco')

    def _sub_comp_amiodarone(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return self._categorical_sub_comp('amiodarone')

    def _sub_comp_fluvastatin(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return self._categorical_sub_comp('fluvastatin')

    def _sub_comp_CYP2C9(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return self._categorical_sub_comp('CYP2C9')

    def _sub_comp_VKORC1(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return self._categorical_sub_comp('VKORC1')

    def _sub_comp_sensitivity(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return self._categorical_sub_comp('sensitivity')

    def _get_history(
            self, list_name: str, length: int) -> Tuple[List[Any], Any, Any]:
        if length == 0:
            raise ValueError('length should be a positive integer, or '
                             '-1 for full length output.')

        if list_name == 'INR':
            _list = self._decision_points_INR_history
            index = self._decision_points_index + 1
            filler = 0.0
            lower, upper = 0.0, 15.0
        elif list_name == 'daily_INR':
            _list = self._full_INR_history
            index = self._day + 1
            filler = 0.0
            lower, upper = 0.0, 15.0
        elif list_name == 'dose':
            _list = self._decision_points_dose_history
            index = self._decision_points_index
            filler = 0.0
            lower = self._action_generator.lower['dose']
            upper = self._action_generator.upper['dose']
        elif list_name == 'daily_dose':
            _list = self._full_dose_history
            index = self._day
            filler = 0.0
            lower = self._action_generator.lower['dose']
            upper = self._action_generator.upper['dose']
        elif list_name == 'interval':
            _list = self._decision_points_interval_history
            index = self._decision_points_index
            filler = 1
            lower = self._action_generator.lower['interval']
            upper = self._action_generator.upper['interval']
        else:
            return [], None, None

        if length == -1:
            result = _list[:index]
        else:
            if length > index:
                i1, i2 = length - index, 0
            else:
                i1, i2 = 0, index-length
            result = [filler] * i1 + _list[i2:index]  # type: ignore

        return result, lower, upper

    def _sub_comp_dose(
            self, _id: int, length: int = 1, **kwargs: Any) -> Dict[str, Any]:
        name = 'dose'
        value, lower, upper = self._get_history(name, length)
        return {'name': name,
                'value': tuple(value),
                'lower': lower,
                'upper': upper}

    def _sub_comp_INR(
            self, _id: int, length: int = 1, **kwargs: Any) -> Dict[str, Any]:
        name = 'INR'
        value, lower, upper = self._get_history(name, length)
        return {'name': name,
                'value': tuple(value),
                'lower': lower,
                'upper': upper}

    def _sub_comp_interval(
            self, _id: int, length: int = 1, **kwargs: Any) -> Dict[str, Any]:
        name = 'interval'
        value, lower, upper = self._get_history(name, length)
        return {'name': name,
                'value': tuple(value),
                'lower': lower,
                'upper': upper}

    def _sub_comp_day(self, _id: int, **kwargs: Any) -> Dict[str, Any]:
        return {'name': 'day',
                'value': self._day if 0 <= self._day < self._max_day else None,
                'lower': 0,
                'upper': self._max_day - 1}

    def _sub_comp_daily_dose(
            self, _id: int, length: int = 1, **kwargs: Any) -> Dict[str, Any]:
        name = 'daily_dose'
        value, lower, upper = self._get_history(name, length)
        return {'name': name,
                'value': tuple(value),
                'lower': lower,
                'upper': upper}

    def _sub_comp_daily_INR(
            self, _id: int, length: int = 1, **kwargs: Any) -> Dict[str, Any]:
        name = 'daily_INR'
        value, lower, upper = self._get_history(name, length)
        return {'name': name,
                'value': tuple(value),
                'lower': lower,
                'upper': upper}

    def _sub_comp_INR_within(
            self, _id: int, length: int = 1, **kwargs: Any) -> Dict[str, Any]:
        name = 'daily_INR'
        l, _, _ = self._get_history('interval', length)
        value, lower, upper = self._get_history(name, sum(l))

        return {'name': name,
                'value': tuple(value),
                'lower': lower,
                'upper': upper}

    def load(self, filename: str, path: Optional[pathlib.Path] = None) -> None:
        '''
        Extends super class's method to make sure 'action_generator' resets if
        it is part of the 'persistent_attributes'.
        '''
        super().load(filename, path)
        if '_action_generator' in self._persistent_attributes:
            self._action_generator.reset()

    def __repr__(self) -> str:
        try:
            temp = ', '.join(''.join(
                (str(k), ': ',
                 ('{:4.2f}' if v.is_numerical else '{}').format(v.value)))
                for k, v in self._patient.feature_set.items())
        except (AttributeError, ValueError, KeyError):
            temp = ''

        return (f'{self.__class__.__qualname__} [{temp}]')

    # if self._action_type == 'dose' and self._current_dose < self._max_dose:
    #     return self._possible_actions[
    #         :int(self._current_dose/self._dose_steps) + 1]
    # elif (self._action_type == 'interval' and
    #         self._current_interval < self._max_interval):
    #     return self._possible_actions[
    #         :int(self._current_interval/self._interval_steps)]
    # elif (self._current_dose < self._max_dose or
    #         self._current_interval < self._max_interval):
    #     self._generate_possible_actions(
    #         dose_cap=self._current_dose, interval_cap=self._current_interval)

    # return self._possible_actions

    # def default_state(self, _id: Optional[int] = None) -> reildata.ReilData:
    #     index = self._decision_points_index - 1
    #     return self._state_normal_fixed + reildata.ReilData([
    #         {'name': 'dose',
    #          'value': (self._decision_points_dose_history[index],),
    #          'lower': self._action_generator.lower['dose'],
    #          'upper': self._action_generator.upper['dose']},
    #         {'name': 'INR',
    #          'value': (self._decision_points_INR_history[index],),
    #          'lower': 0.0,
    #          'upper': 15.0},
    #         {'name': 'interval',
    #          'value': (self._decision_points_interval_history[index],),
    #          'lower': self._action_generator.lower['interval'],
    #          'upper': self._action_generator.upper['interval']}],
    #         lazy_evaluation=True)


# def stats(self, stats_list: Tuple[str, ...]) -> Dict[str, Any]:
#     results = {}
#     for s in stats_list:
#         if s == 'TTR':
#             temp = reil_functions.Functions.TTR(self._full_INR_history)
#             # INR = self._full_INR_history
#             # sum(
#             #     (1 if 2.0 <= INRi <= 3.0 else 0 for INRi in INR)
#             #       ) / len(INR)
#         elif s == 'dose_change':
#             temp = reil_functions.Functions.dose_change_count(
#                 self._full_dose_history)
#             # temp = sum(x != self._full_dose_history[i+1]
#             #            for i, x in enumerate(self._full_dose_history[:-1]))
#         elif s == 'delta_dose':
#             temp = reil_functions.Functions.delta_dose(
#                 self._full_dose_history)
#             # temp = sum(abs(x-self._full_dose_history[i+1])
#             #            for i, x in enumerate(self._full_dose_history[:-1]))
#         else:
#             self._logger.warning(
#                 f'WARNING! {s} is not one of the available stats!')
#             continue

#         results[s] = temp

#     results['ID'] = reildata.ReilData([
#         {'name': 'age',
#          'value': self._patient.feature_set['age'].value,
#          'lower': self._patient.feature_set['age'].lower,
#          'upper': self._patient.feature_set['age'].upper},
#         {'name': 'CYP2C9',
#          'value': self._patient.feature_set['CYP2C9'].value,
#          'categories': self._patient.feature_set['CYP2C9'].categories},
#         {'name': 'VKORC1',
#          'value': self._patient.feature_set['VKORC1'].value,
#          'categories': self._patient.feature_set['VKORC1'].categories}],
#         lazy_evaluation=True)

#     return results

# @property
# def _INR_history(self) -> List[float]:
#     # INR has one more value (initial INR) compared to dose.
#     return [0.0]*(self._INR_history_length - self._decision_points_index) \
#         + self._decision_points_INR_history[
#             :self._decision_points_index + 1][-self._INR_history_length - 1:]

# @property
# def _dose_history(self) -> List[float]:
#     return [0.0]*(self._dose_history_length - self._decision_points_index) \
#         + self._decision_points_dose_history[
#             :self._decision_points_index][-self._dose_history_length:] \
#         + ([self._current_dose]
#            if self._action_type == 'interval_only' else [])

# @property
# def _interval_history(self) -> List[int]:
#     return [0]*(
#           self._interval_history_length - self._decision_points_index) \
#         + self._decision_points_interval_history[
#             :self._decision_points_index][-self._interval_history_length:] \
#         + ([self._current_interval]
#            if self._action_type == 'dose_only' else [])

# @property
# def _interval_history_length(self) -> int:
#     return max(self._dose_history_length, self._INR_history_length)

# @property
# def state(self) -> reildata.ReilData:
#     if self._ex_protocol_current['state'] == 'extended':
#         return self._state_extended()
#     else:
#         return self._state_normal()

# def _prepare_reward_arguments(self,
#                               arguments: Tuple[str, ...],
#                               observation_length: int,
#                               retrospective: bool,
#                               interpolate: bool) -> Dict[str, Any]:
#     output = {'y': [], 'x': []}
#     if retrospective:
#         if interpolate:
#             if observation_length == -1:
#                 start = 0
#             else:
#                 start = self._decision_points_index - observation_length + 1
#             end = self._decision_points_index + 1
#             if 'INR' in arguments:
#                 output = {
#                     'y': self._decision_points_INR_history[start:end],
#                     'x': self._decision_points_interval_history[start:end-1]}
#             elif 'Doses' in arguments:
#                 output = {
#                     'y': self._decision_points_dose_history[start:end],
#                     'x': self._decision_points_interval_history[start:end-1]}
#         else:
#             if observation_length == -1:
#                 start = 0
#             else:
#                 start = self._day - observation_length + 1
#             end = self._day + 1
#             if 'INR' in arguments:
#                 output = {'y': self._full_INR_history[start:end]}
#             elif 'Doses' in arguments:
#                 output = {'y': self._full_dose_history[start:end]}

#     else:
#         start = self._day + 1
#         if observation_length == -1:
#             end = self._max_day
#         else:
#             end = min(self._day + observation_length + 1, self._max_day)

#         current_dose = self._decision_points_dose_history[
#             self._decision_points_index]

#         if interpolate:
#             if 'INR' in arguments:
#                 temp_patient = copy.deepcopy(self._patient)
#                 INRs_temp = temp_patient.model(
#                     dose=dict((i, current_dose)
#                               for i in range(start, end)),
#                     measurement_days=[start, end])['INR']
#                 output = {'y': INRs_temp,
#                           'x': [start, end]}
#             elif 'Doses' in arguments:
#                 output = {'y': [current_dose] * 2,
#                           'x': [start, end]}
#         else:
#             if 'INR' in arguments:
#                 temp_patient = copy.deepcopy(self._patient)
#                 INRs_temp = temp_patient.model(
#                     dose=dict((i, current_dose)
#                               for i in range(start, end)),
#                     measurement_days=list(range(start, end)))['INR']
#                 output = {'y': INRs_temp}
#             elif 'Doses' in arguments:
#                 output = {'y': [current_dose] * (end - start)}

#     return output

# def _state_extended(self) -> reildata.ReilData:
#     return self._state_extended_fixed + reildata.ReilData([
#         {'name': 'day',
#          'value': self._day if 0 <= self._day < self._max_day else None,
#          'lower': 0,
#          'upper': self._max_day - 1},
#         {'name': 'Doses',
#          'value': tuple(self._dose_history),
#          'lower': self._action_generator.lower['dose'],
#          'upper': self._action_generator.upper['dose']},
#         {'name': 'INR',
#          'value': tuple(self._INR_history),
#          'lower': 0.0,
#          'upper': 15.0},
#         {'name': 'Intervals',
#          'value': tuple(self._interval_history),
#          'lower': self._action_generator.lower['interval'],
#          'upper': self._action_generator.upper['interval']}],
#         lazy_evaluation=True)

# def _state_normal(self) -> reildata.ReilData:
#     return self._state_normal_fixed + reildata.ReilData([
#         {'name': 'Doses',
#          'value': tuple(self._dose_history),
#          'lower': self._action_generator.lower['dose'],
#          'upper': self._action_generator.upper['dose']},
#         {'name': 'INR',
#          'value': tuple(self._INR_history),
#          'lower': 0.0,
#          'upper': 15.0},
#         {'name': 'Intervals',
#          'value': tuple(self._interval_history),
#          'lower': self._action_generator.lower['interval'],
#          'upper': self._action_generator.upper['interval']}],
#         lazy_evaluation=True)



# -*- coding: utf-8 -*-
'''
cancer_model class
==================

This `cancer_model` class implements a four-state nonlinear cancer
chemotherapy model proposed by Lisette et al. (2003).
'''

from reil.subjects.healthcare import Patient
from reil.datatypes import Feature
from reil.utils import functions


class CancerPatient(Patient):
    feature_set = {
        'age': Feature.numerical(  # Aurora population
            lower=18.0, upper=100.0, mean=67.30, stdev=13.43,
            generator=functions.random_truncated_normal),
        'weight': Feature.numerical(  # lb  - Aurora population
            lower=70.0, upper=500.0, mean=199.24, stdev=54.71,
            generator=functions.random_truncated_normal),
        'height': Feature.numerical(  # in - Aurora population
            lower=45.0, upper=85.0, mean=66.78, stdev=4.31,
            generator=functions.random_truncated_normal),
        'gender': Feature.categorical(  # Aurora population
            categories=('Female', 'Male'),
            probabilities=(0.5314, 0.4686),
            generator=functions.random_categorical)}

    '''
    Four-state nonlinear cancer chemotherapy model.
    
    Attributes
    ----------
        state: the state of the subject as a ValueSet.
        is_terminated: whether the subject is finished or not.
        possible_actions: a list of possible actions.

    Methods
    -------
        register: register a new agent and return its ID or return ID of an existing agent.
        take_effect: get an action and change the state accordingly.
        reset: reset the state and is_terminated.
    '''
    def __init__(self, **kwargs):
        self.set_defaults(
            drug={'initial_value': 0, 'infusion_rate': 0, 'decay_rate': 0,
                'normal_cell_kill_rate': 0, 'tumor_cell_kill_rate': 0, 'immune_cell_kill_rate': 0},
            normal_cells={'initial_value': 0, 'growth_rate': 0, 'carrying_capacity': 0},
            tumor_cells={'initial_value': 0, 'growth_rate': 0, 'carrying_capacity': 0},
            immune_cells={'initial_value': 0, 'influx_rate': 0, 'threshold_rate': 1, 'response_rate': 0, 'death_rate': 0},
            competition_term={'normal_from_tumor':0, 'tumor_from_normal': 0, 'tumor_from_immune': 0, 'immune_from_tumor': 0},
            x={'drug': 0, 'normal_cells': 0, 'tumor_cells': 0, 'immune_cells': 0, 'day': 1},
            state_function=lambda x: {'value': 0, 'min':0, 'max':0}, rl_function=lambda new_x, old_x: 1-new_x['cancer']/old_x['cancer'],
            termination_check= lambda x: x['tumor_cells']==0,
            e=lambda x: 0, state_range=[0],
            u_max=10, u_steps=20)
        self.set_params(**kwargs)
        super().__init__(**kwargs)


        self._x = {'drug': self._drug['initial_value'], 'normal_cells': self._normal_cells['initial_value'],
                'tumor_cells': self._tumor_cells['initial_value'], 'immune_cells': self._immune_cells['initial_value'], 'day': 1}

    @property
    def state(self):
        return ValueSet(**self._state_function(self._x))
        e = self._e(self._x)
        for i, _ in enumerate(self._state_range):
            if e <= self._state_range[i]:
                return ValueSet(i, min=0, max=len(self._state_range))
        return ValueSet(len(self._state_range), min=0, max=len(self._state_range))

    @property
    def is_terminated(self):
        return self._termination_check(self._x)

    @property
    def possible_actions(self):
        return ValueSet([self._u_max*x/self._u_steps for x in range(0, self._u_steps+1)], min=0, max=self._u_max, 
                        binary=lambda x: (int(x * self._u_steps // self._u_max), self._u_steps+1)).as_valueset_array()

    def register(self, agent_name):
        '''
        Registers an agent and returns its ID. If the agent is new, a new ID is generated and the agent_name is added to agent_list.
        \nArguments:
        \n    agent_name: the name of the agent to be registered.
        '''
        try:
            return self._agent_list[agent_name]
        except KeyError:
            if len(self._agent_list) == 1:
                raise ValueError('Only one drug is allowed.')
            self._agent_list[agent_name] = 1
            return 1

    def take_effect(self, action, _id=None):
        self._drug['infusion_rate'] = action.value[0]
        x_dot = {}
        x_dot['day'] = 1
        x_dot['normal_cells'] = self._x['normal_cells'] * (
                                    self._normal_cells['growth_rate'] * (1 - self._normal_cells['carrying_capacity']*self._x['normal_cells'])
                                    - self._competition_term['normal_from_tumor'] * self._x['tumor_cells']
                                    - self._drug['normal_cell_kill_rate'] * (1-exp(-self._x['drug'])))

        x_dot['tumor_cells'] = self._x['tumor_cells'] * (
                                    self._tumor_cells['growth_rate'] * (1 - self._tumor_cells['carrying_capacity']*self._x['tumor_cells'])
                                    - self._competition_term['tumor_from_immune'] * self._x['immune_cells']
                                    - self._competition_term['tumor_from_normal'] * self._x['normal_cells']
                                    - self._drug['tumor_cell_kill_rate'] * (1-exp(-self._x['drug'])))

        x_dot['immune_cells'] = self._x['immune_cells'] * (
                                    self._immune_cells['response_rate']*self._x['tumor_cells']/(self._immune_cells['threshold_rate']+self._x['tumor_cells'])
                                    - self._immune_cells['death_rate']
                                    - self._competition_term['immune_from_tumor'] * self._x['tumor_cells']
                                    - self._drug['immune_cell_kill_rate'] * (1-exp(-self._x['drug']))) + self._immune_cells['influx_rate']

        x_dot['drug'] = self._x['drug'] * (-self._drug['decay_rate']) + self._drug['infusion_rate']
        new_x = {}
        for i in self._x:
            try:
                new_x[i] = max(self._x[i] + x_dot[i], 0)  # I manually enforced >=0 constraint, but it shouldn't be!
            except:
                new_x[i] = self._x[i]

        r = self._rl_function(new_x, self._x)
        self._x = new_x

        return r

    def reset(self):
        self._x = {'drug': self._drug['initial_value'], 'normal_cells': self._normal_cells['initial_value'],
                'tumor_cells': self._tumor_cells['initial_value'], 'immune_cells': self._immune_cells['initial_value'], 'day': 1}

    def __repr__(self):
        try:
            return f"CancerModel: [day: {self._x['day']}, N: {self._x['normal_cells']}, T: {self._x['tumor_cells']}, N: {self._x['immune_cells']}, C: {self._x['drug']}]"
        except:
            return 'CancerModel'



# -*- coding: utf-8 -*-
'''
functions module
================

Contains some useful functions.
'''

import math
import random
from typing import Iterable, Tuple

import numpy as np
from reil.datatypes import Feature
from scipy import stats


def random_truncated_normal(f: Feature) -> float:
    return min(max(
        np.random.normal(f.mean, f.stdev), f.lower),
        f.upper)


def random_uniform(f: Feature):
    return np.random.uniform(f.lower, f.upper)


def random_categorical(f: Feature):
    if f.probabilities is None:
        return random.choice(f.categories)  # type:ignore
    else:
        return np.random.choice(
            f.categories, 1, p=f.probabilities)[0]


def random_truncated_lnorm(f: Feature) -> float:
    # capture 50% of the data.
    # This restricts the log values to a "reasonable" range
    quartileRange = (0.25, 0.75)
    lnorm = stats.lognorm(f.stdev, scale=math.exp(f.mean))  # type:ignore
    qValues = lnorm.ppf(quartileRange)
    values = list(v for v in lnorm.rvs(size=1000)
                  if (v > qValues[0]) & (v < qValues[1]))
    return random.sample(values, 1)[0]


def square_dist(x: float, y: Iterable[float]) -> float:
    return sum((x - yi) ** 2
               for yi in y)


def in_range(r: Tuple[float, float], x: Iterable[float]) -> int:
    return sum(r[0] <= xi <= r[1]
               for xi in x)


def interpolate(start: float, end: float, steps: int) -> Iterable[float]:
    return (start + (end - start) / steps * j
            for j in range(1, steps + 1))



# -*- coding: utf-8 -*-
'''
UserAgent class
===============

An agent that prints the state and asks the user for action.
'''
from typing import Any, Optional, Tuple

from reil import agents
from reil.datatypes.reildata import ReilData


class UserAgent(agents.NoLearnAgent):
    '''
    An agent that acts based on user input.
    '''

    def __init__(self,
                 default_actions: Tuple[ReilData, ...] = (),
                 **kwargs: Any):
        super().__init__(default_actions=default_actions, **kwargs)

    def act(self,
            state: ReilData,
            actions: Optional[Tuple[ReilData, ...]] = None,
            epoch: int = 0) -> ReilData:
        '''
        Return a random action.

        Arguments
        ---------
        state:
            The state for which the action should be returned.

        actions:
            The set of possible actions to choose from.

        epoch:
            The epoch in which the agent is acting.

        Returns
        -------
        :
            The action
        '''
        possible_actions = actions or self._default_actions

        action = None
        while action is None:
            for i, a in enumerate(possible_actions):
                print(f'{i}. {a.value}')
            action = int(input(
                f'Choose action number for this state: {state.value}'))

        return possible_actions[action]
